"""Test functionalities of cobra component validation functions."""

import pytest

from cobra.core import Metabolite, Model, Reaction
from cobra.manipulation import check_mass_balance, check_metabolite_compartment_formula


def test_validate_mass_balance(model: Model) -> None:
    """Test reaction mass balance validation."""
    assert len(check_mass_balance(model)) == 0
    # if we remove the SBO term which marks the reaction as
    # mass balanced, then the reaction should be detected as
    # no longer mass balanced
    EX_rxn = model.reactions.query(lambda r: r.boundary)[0]
    EX_rxn.annotation.pop("sbo")
    balance = check_mass_balance(model)
    assert len(balance) == 1
    assert EX_rxn in balance
    m1 = Metabolite("m1", formula="()")
    r1 = Reaction("r1")
    r1.add_metabolites({m1: 1})
    with pytest.raises(ValueError), pytest.warns(UserWarning):
        r1.check_mass_balance()


def test_validate_formula_compartment(model: Model) -> None:
    """Test metabolite formulae validation."""
    model.metabolites[1].formula = "(a*.bcde)"
    errors = check_metabolite_compartment_formula(model)
    assert len(errors) == 1



"""Provide the base class and utility function for gap filling."""


import logging
from typing import TYPE_CHECKING, Dict, List, Optional

from optlang.interface import OPTIMAL
from optlang.symbolics import Zero

from ..core import Model
from ..util import fix_objective_as_constraint, interface_to_str


if TYPE_CHECKING:
    from cobra import Reaction


logger = logging.getLogger(__name__)


class GapFiller:
    r"""
    The base class for performing gap filling.

    This class implements gap filling based on a mixed-integer approach,
    very similar to that described in [1]_ and the 'no-growth but growth'
    part of [2]_ but with minor adjustments. In short, we add indicator
    variables for using the reactions in the universal model, z_i and then
    solve problem

    minimize: \sum_i c_i * z_i
    s.t.    : Sv = 0
              v_o \ge t
              lb_i \le v_i \le ub_i
              v_i = 0 if z_i = 0

    where lb, ub are respectively the upper, lower flux bounds for reaction i,
    c_i is a cost parameter and the objective v_o is greater than the lower
    bound t. The default costs are 1 for reactions from the universal model,
    100 for exchange (uptake) reactions added and 1 for added demand reactions.

    Note that this is a mixed-integer linear program and as such will be
    expensive to solve for large models. Consider using alternatives [3]_
    such as CORDA instead [4,5]_ .

    Parameters
    ----------
    model : cobra.Model
        The model to perform gap filling on.
    universal : cobra.Model, optional
        A universal model with reactions that can be used to complete the
        `model` (default None).
    lower_bound : float, optional
        The minimally accepted flux for the objective in the filled model
        (default 0.05).
    penalties : dict, optional
        A dictionary with keys being 'universal' (all reactions included in
        the universal model), 'exchange' and 'demand' (all additionally
        added exchange and demand reactions) for the three reaction types.
        Can also have reaction identifiers for reaction specific costs.
        Defaults are 1, 100 and 1 respectively (default None).
    exchange_reactions : bool, optional
        Consider adding exchange (uptake) reactions for all metabolites
        in the model (default False).
    demand_reactions : bool, optional
        Consider adding demand reactions for all metabolites (default True).
    integer_threshold : float, optional
        The threshold at which a value is considered non-zero (aka
        integrality threshold). If gapfilled models fail to validate,
        you may want to lower this value (default 1E-6).

    Attributes
    ----------
    indicators: list of optlang.interface.Variable
        The list of symbolic indicator variables.
    costs: dict of {optlang.interface.Variable: float}
        The dictionary with symbolic variables as keys and their cost as
        values.

    References
    ----------
    .. [1] Reed, Jennifer L., Trina R. Patel, Keri H. Chen, Andrew R. Joyce,
       Margaret K. Applebee, Christopher D. Herring, Olivia T. Bui, Eric M.
       Knight, Stephen S. Fong, and Bernhard O. Palsson. “Systems Approach
       to Refining Genome Annotation.” Proceedings of the National Academy
       of Sciences 103, no. 46 (2006): 17480–17484.

    .. [2] Kumar, Vinay Satish, and Costas D. Maranas. “GrowMatch: An
       Automated Method for Reconciling In Silico/In Vivo Growth
       Predictions.” Edited by Christos A. Ouzounis. PLoS Computational
       Biology 5, no. 3 (March 13, 2009): e1000308.
       doi:10.1371/journal.pcbi.1000308.

    .. [3] http://opencobra.github.io/cobrapy/tags/gapfilling/

    .. [4] Schultz, André, and Amina A. Qutub. “Reconstruction of
       Tissue-Specific Metabolic Networks Using CORDA.” Edited by Costas D.
       Maranas. PLOS Computational Biology 12, no. 3 (March 4, 2016):
       e1004808. doi:10.1371/journal.pcbi.1004808.

    .. [5] Diener, Christian https://github.com/cdiener/corda

    """

    def __init__(
        self,
        model: Model,
        universal: Optional[Model] = None,
        lower_bound: float = 0.05,
        penalties: Optional[Dict[str, "Reaction"]] = None,
        exchange_reactions: bool = False,
        demand_reactions: bool = True,
        integer_threshold: float = 1e-6,
        **kwargs,
    ) -> None:
        """Initialize a new GapFiller object.

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        """
        self.original_model = model
        self.lower_bound = lower_bound
        self.model = model.copy()
        tolerances = self.model.solver.configuration.tolerances
        try:
            tolerances.integrality = integer_threshold
        except AttributeError:
            logger.warning(
                f"The current solver interface {interface_to_str(self.model.problem)} "
                f"doesn't support setting the integrality tolerance."
            )
        # TODO (Midnighter): One could debate how useful it is to compare against this
        #  threshold when it is not supported by the chosen solver.
        self.integer_threshold = integer_threshold
        self.universal = universal.copy() if universal else Model("universal")
        self.penalties = dict(universal=1, exchange=100, demand=1)
        if penalties is not None:
            self.penalties.update(penalties)
        self.indicators = []
        self.costs = {}
        self.extend_model(exchange_reactions, demand_reactions)
        fix_objective_as_constraint(self.model, bound=lower_bound)
        self.add_switches_and_objective()

    def extend_model(
        self, exchange_reactions: bool = False, demand_reactions: bool = True
    ) -> None:
        """Extend gap filling model.

        Add reactions from universal model and optionally exchange and
        demand reactions for all metabolites in the model to perform
        gap filling on.

        Parameters
        ----------
        exchange_reactions : bool, optional
            Consider adding exchange (uptake) reactions for all metabolites
            in the model (default False).
        demand_reactions : bool, optional
            Consider adding demand reactions for all metabolites
            (default True).

        """
        for rxn in self.universal.reactions:
            rxn.gapfilling_type = "universal"
        new_metabolites = self.universal.metabolites.query(
            lambda metabolite: metabolite not in self.model.metabolites
        )
        self.model.add_metabolites(new_metabolites)
        existing_exchanges = []
        for rxn in self.universal.boundary:
            existing_exchanges = existing_exchanges + [
                met.id for met in list(rxn.metabolites)
            ]

        for met in self.model.metabolites:
            if exchange_reactions:
                # check for exchange reaction in model already
                if met.id not in existing_exchanges:
                    rxn = self.universal.add_boundary(
                        met,
                        type="exchange_smiley",
                        lb=-1000,
                        ub=0,
                        reaction_id=f"EX_{met.id}",
                    )
                    rxn.gapfilling_type = "exchange"
            if demand_reactions:
                rxn = self.universal.add_boundary(
                    met,
                    type="demand_smiley",
                    lb=0,
                    ub=1000,
                    reaction_id=f"DM_{met.id}",
                )
                rxn.gapfilling_type = "demand"

        new_reactions = self.universal.reactions.query(
            lambda reaction: reaction not in self.model.reactions
        )
        self.model.add_reactions(new_reactions)

    def update_costs(self) -> None:
        """Update coefficients for the indicator variables in the objective.

        Done incrementally so that second time the function is called,
        active indicators in the current solutions gets higher cost than the
        unused indicators.

        """
        for var in self.indicators:
            if var not in self.costs:
                self.costs[var] = var.cost
            else:
                if var._get_primal() > self.integer_threshold:
                    self.costs[var] += var.cost
        self.model.objective.set_linear_coefficients(self.costs)

    def add_switches_and_objective(self) -> None:
        """Update gap filling model with switches and indicator objective."""
        constraints = []
        big_m = max(max(abs(b) for b in r.bounds) for r in self.model.reactions)
        prob = self.model.problem
        for rxn in self.model.reactions:
            if not hasattr(rxn, "gapfilling_type"):
                continue
            indicator = prob.Variable(
                name=f"indicator_{rxn.id}", lb=0, ub=1, type="binary"
            )
            if rxn.id in self.penalties:
                indicator.cost = self.penalties[rxn.id]
            else:
                indicator.cost = self.penalties[rxn.gapfilling_type]
            indicator.rxn_id = rxn.id
            self.indicators.append(indicator)

            # if z = 1 v_i is allowed non-zero
            # v_i - Mz <= 0   and   v_i + Mz >= 0
            constraint_lb = prob.Constraint(
                rxn.flux_expression - big_m * indicator,
                ub=0,
                name=f"constraint_lb_{rxn.id}",
                sloppy=True,
            )
            constraint_ub = prob.Constraint(
                rxn.flux_expression + big_m * indicator,
                lb=0,
                name=f"constraint_ub_{rxn.id}",
                sloppy=True,
            )

            constraints.extend([constraint_lb, constraint_ub])

        self.model.add_cons_vars(self.indicators)
        self.model.add_cons_vars(constraints, sloppy=True)
        self.model.objective = prob.Objective(Zero, direction="min", sloppy=True)
        self.model.objective.set_linear_coefficients({i: 1 for i in self.indicators})
        self.update_costs()

    def fill(self, iterations: int = 1) -> List[List["Reaction"]]:
        """Perform the gap filling.

        With every iteration, it solves the model, updates the costs and
        records the used reactions.

        Parameters
        ----------
        iterations : int, optional
            The number of rounds of gap filling to perform. For every
            iteration, the penalty for every used reaction increases
            linearly. This way, the algorithm is encouraged to search for
            alternative solutions which may include previously used
            reactions i.e., with enough iterations pathways including 10
            steps will eventually be reported even if the shortest pathway
            is a single reaction (default 1).

        Returns
        -------
        list of list of cobra.Reaction
            A list of lists where each element is a list of reactions that
            were used to gap fill the model.

        Raises
        ------
        RuntimeError
            If the model fails to be validated (i.e. the original model with
            the proposed reactions added, still cannot get the required flux
            through the objective).

        """
        used_reactions = []
        for _ in range(iterations):
            self.model.slim_optimize(
                error_value=None, message="gap filling optimization failed"
            )
            solution = [
                self.model.reactions.get_by_id(ind.rxn_id)
                for ind in self.indicators
                if ind._get_primal() > self.integer_threshold
            ]
            if not self.validate(solution):
                raise RuntimeError(
                    "Failed to validate gap filled model, "
                    "try lowering the integer threshold."
                )
            used_reactions.append(solution)
            self.update_costs()
        return used_reactions

    def validate(self, reactions: List["Reaction"]) -> bool:
        """Validate the model.

        Parameters
        ----------
        reactions: list of cobra.Reaction
            The reactions to add to the model for validation.

        Returns
        -------
        bool
            Whether the model is valid or not.

        """
        with self.original_model as model:
            mets = [x.metabolites for x in reactions]
            all_keys = set().union(*(d.keys() for d in mets))
            model.add_metabolites(all_keys)
            model.add_reactions(reactions)
            model.slim_optimize()
            return (
                model.solver.status == OPTIMAL
                and model.solver.objective.value >= self.lower_bound
            )


def gapfill(
    model: Model,
    universal: Optional[Model] = None,
    lower_bound: float = 0.05,
    penalties: Optional[Dict[str, "Reaction"]] = None,
    demand_reactions: bool = True,
    exchange_reactions: bool = False,
    iterations: int = 1,
):
    """Perform gap filling on a model.

    Parameters
    ----------
    model : cobra.Model
        The model to perform gap filling on.
    universal : cobra.Model, optional
        A universal model with reactions that can be used to complete the
        model. Only gapfill considering demand and exchange reactions if
        left missing (default None).
    lower_bound : float, optional
        The minimally accepted flux for the objective in the filled model.
        (default 0.05).
    penalties : dict, optional
        A dictionary with keys being 'universal' (all reactions included in
        the universal model), 'exchange' and 'demand' (all additionally
        added exchange and demand reactions) for the three reaction types.
        Can also have reaction identifiers for reaction specific costs.
        Defaults are 1, 100 and 1 respectively (default None).
    exchange_reactions : bool, optional
        Consider adding exchange (uptake) reactions for all metabolites
        in the model (default False).
    demand_reactions : bool, optional
        Consider adding demand reactions for all metabolites (default True).
    iterations : int, optional
        The number of rounds of gap filling to perform. For every iteration,
        the penalty for every used reaction increases linearly. This way,
        the algorithm is encouraged to search for alternative solutions
        which may include previously used reactions i.e., with enough
        iterations pathways including 10 steps will eventually be reported
        even if the shortest pathway is a single reaction (default 1).

    Returns
    -------
    list of list of cobra.Reaction
        A list of lists with on set of reactions that completes the model per
        requested iteration.

    Examples
    --------
    >>> import cobra.test as ct
    >>> from cobra import Model
    >>> from cobra.flux_analysis import gapfill
    >>> model = ct.create_test_model("salmonella")
    >>> universal = Model("universal")
    >>> universal.add_reactions([model.reactions.GF6PTA.copy()])
    >>> model.remove_reactions([model.reactions.GF6PTA])
    >>> gapfill(model, universal)
    [[<Reaction GF6PTA at 0x12206a280>]]

    """
    gapfiller = GapFiller(
        model,
        universal=universal,
        lower_bound=lower_bound,
        penalties=penalties,
        demand_reactions=demand_reactions,
        exchange_reactions=exchange_reactions,
    )
    return gapfiller.fill(iterations=iterations)



"""Provide functions to identify the type of boundary reactions.

This module uses various heuristics to decide whether a boundary reaction
is an exchange, demand or sink reaction. It mostly orientates on the
following paper:

Thiele, I., & Palsson, B. Ø. (2010, January). A protocol for
generating a high-quality genome-scale metabolic reconstruction.
Nature protocols. Nature Publishing Group.
http://doi.org/10.1038/nprot.2009.203

"""

import logging
from typing import TYPE_CHECKING, List, Optional

import pandas as pd

from .annotations import compartment_shortlist, excludes, sbo_terms


if TYPE_CHECKING:
    from cobra import Model, Reaction


logger = logging.getLogger(__name__)


def find_external_compartment(model: "Model") -> str:
    """Find the external compartment in the model.

    Uses a simple heuristic where the external compartment should be the
    one with the most exchange reactions.

    Parameters
    ----------
    model : cobra.Model
        The cobra model whose external compartments are to be identified.

    Returns
    -------
    str
        The putative external compartment.

    Raises
    ------
    RuntimeError
        If several compartments are similar and thus difficult to identify,
        or, recognized names usually used for external compartment are
        absent.

    """
    if model.boundary:
        counts = pd.Series(tuple(r.compartments)[0] for r in model.boundary)
        most = counts.value_counts()
        most = most.index[most == most.max()].to_series()
    else:
        most = None
    like_external = compartment_shortlist["e"] + ["e"]
    matches = pd.Series(
        [co in like_external for co in model.compartments], index=model.compartments
    )

    if matches.sum() == 1:
        compartment = matches.index[matches][0]
        logger.info(
            f"Compartment `{compartment}` sounds like an external compartment. "
            "Using this one without counting boundary reactions."
        )
        return compartment
    elif most is not None and matches.sum() > 1 and matches[most].sum() == 1:
        compartment = most[matches[most]][0]
        logger.warning(
            "There are several compartments that look like an "
            f"external compartment but `{compartment}` has the most boundary "
            "reactions, so using that as the external compartment."
        )
        return compartment
    elif matches.sum() > 1:
        raise RuntimeError(
            "There are several compartments that look "
            "like external compartments but we can't tell "
            "which one to use. Consider renaming your "
            "compartments please."
        )

    if most is not None:
        return most[0]
        logger.warning(
            "Could not identify an external compartment by name and "
            "choosing one with the most boundary reactions. That "
            "might be complete nonsense or change suddenly. "
            "Consider renaming your compartments using "
            "`Model.compartments` to fix this."
        )
    # No info in the model, so give up
    raise RuntimeError(
        "The heuristic for discovering an external compartment "
        "relies on names and boundary reactions. Yet, there "
        "are neither compartments with recognized names nor "
        "boundary reactions in the model."
    )


def is_boundary_type(
    reaction: "Reaction", boundary_type: str, external_compartment: str
) -> bool:
    """Check whether a reaction is an exchange reaction.

    Parameters
    ----------
    reaction : cobra.Reaction
        The reaction to check.
    boundary_type : {"exchange", "demand", "sink"}
        Boundary type to check for.
    external_compartment : str
        The ID for the external compartment.

    Returns
    -------
    bool
        Whether the reaction looks like the requested type. Might be based
        on a heuristic.

    """
    # Check if the reaction has an annotation. Annotations dominate everything.
    sbo_term = reaction.annotation.get("sbo", "")
    if isinstance(sbo_term, list):
        sbo_term = sbo_term[0]
    sbo_term = sbo_term.upper()

    if sbo_term == sbo_terms[boundary_type]:
        return True
    if sbo_term in [sbo_terms[k] for k in sbo_terms if k != boundary_type]:
        return False

    # Check if the reaction is in the correct compartment (exterior or inside)
    correct_compartment = external_compartment in reaction.compartments
    if boundary_type != "exchange":
        correct_compartment = not correct_compartment

    # Check if the reaction has the correct reversibility
    rev_type = True
    if boundary_type == "demand":
        rev_type = not reaction.reversibility
    elif boundary_type == "sink":
        rev_type = reaction.reversibility

    return (
        reaction.boundary
        and not any(ex in reaction.id for ex in excludes[boundary_type])
        and correct_compartment
        and rev_type
    )


def find_boundary_types(
    model: "Model", boundary_type: str, external_compartment: Optional[str] = None
) -> List["Reaction"]:
    """Find specific boundary reactions.

    Parameters
    ----------
    model : cobra.Model
        The cobra model whose boundary reactions are to be found.
    boundary_type : {"exchange", "demand", "sink"}
        Boundary type to check for.
    external_compartment : str, optional
        The ID for the external compartment. If None, it will be detected
        automatically (default None).

    Returns
    -------
    list of cobra.Reaction or an empty list
        A list of likely boundary reactions of a user defined type.

    """
    if not model.boundary:
        logger.warning(
            "There are no boundary reactions in this model. "
            "Therefore, specific types of boundary reactions such "
            "as 'exchanges', 'demands' or 'sinks' cannot be "
            "identified."
        )
        return []
    if external_compartment is None:
        external_compartment = find_external_compartment(model)
    return model.reactions.query(
        lambda r: is_boundary_type(r, boundary_type, external_compartment)
    )



# -*- coding: utf-8 -*-

"""Define the singleton meta class."""

from __future__ import absolute_import


class Singleton(type):
    """Implementation of the singleton pattern as a meta class."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Override an inheriting class' call."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



"""Provide a concrete implementation of the BioModels repository interface."""


from io import BytesIO

import httpx

from .abstract_model_repository import AbstractModelRepository


class BiGGModels(AbstractModelRepository):
    """
    Define a concrete implementation of the BiGG Models repository.

    Attributes
    ----------
    name : str
        The name of the BiGG Models repository.

    """

    name: str = "BiGG Models"

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """
        Initialize a BiGG Models repository interface.

        Other Parameters
        ----------------
        kwargs
            Passed to the parent constructor in order to enable multiple inheritance.

        """
        super().__init__(url="http://bigg.ucsd.edu/static/models/", **kwargs)

    def get_sbml(self, model_id: str) -> bytes:
        """
        Attempt to download an SBML document from the repository.

        Parameters
        ----------
        model_id : str
            The identifier of the desired metabolic model. This is typically repository
            specific.

        Returns
        -------
        bytes
            A gzip-compressed, UTF-8 encoded SBML document.

        Raises
        ------
        httpx.HTTPError
            In case there are any connection problems.

        """
        compressed = BytesIO()
        filename = f"{model_id}.xml.gz"
        with self._progress, httpx.stream(
            method="GET", url=self._url.join(filename)
        ) as response:
            response.raise_for_status()
            task_id = self._progress.add_task(
                description="download",
                total=int(response.headers["Content-Length"]),
                model_id=model_id,
            )
            for chunk in response.iter_bytes():
                compressed.write(chunk)
                self._progress.update(task_id=task_id, advance=len(chunk))
        compressed.seek(0)
        return compressed.read()



"""Provide the reaction summary class."""


import logging
from textwrap import dedent, shorten
from typing import TYPE_CHECKING, Optional, Union

import pandas as pd

from cobra.flux_analysis import flux_variability_analysis, pfba
from cobra.summary import Summary


if TYPE_CHECKING:
    from cobra import Model, Reaction, Solution


logger = logging.getLogger(__name__)


class ReactionSummary(Summary):
    """
    Define the reaction summary.

    See Also
    --------
    Summary : Parent that defines further attributes.
    MetaboliteSummary
    ModelSummary

    """

    def __init__(
        self,
        *,
        reaction: "Reaction",
        model: "Model",
        solution: Optional["Solution"] = None,
        fva: Optional[Union[float, pd.DataFrame]] = None,
        **kwargs,
    ) -> None:
        """
        Initialize a reaction summary.

        Parameters
        ----------
        reaction: cobra.Reaction
            The reaction object whose summary we intend to get.
        model : cobra.Model
            The metabolic model in which to generate a reaction summary.
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution (default None).
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched (default None).

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        See Also
        --------
        Summary : Parent that has further default parameters.
        MetaboliteSummary
        ModelSummary

        """
        super(ReactionSummary, self).__init__(**kwargs)
        self._reaction = reaction.copy()
        self._generate(model, solution, fva)

    def _generate(
        self,
        model: "Model",
        solution: Optional["Solution"],
        fva: Optional[Union[float, pd.DataFrame]],
    ) -> None:
        """
        Prepare the data for the summary instance.

        Parameters
        ----------
        model : cobra.Model
            The metabolic model for which to generate a metabolite summary.
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution.
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched.

        """
        super()._generate(model=model, solution=solution, fva=fva)

        if solution is None:
            logger.info("Generating new parsimonious flux distribution.")
            solution = pfba(model)

        if isinstance(fva, float):
            logger.info("Performing flux variability analysis.")
            fva = flux_variability_analysis(
                model,
                reaction_list=[self._reaction],
                fraction_of_optimum=fva,
            )

        # Create the basic flux table.
        self._flux = pd.DataFrame(
            data={"flux": [solution[self._reaction.id]]},
            index=[self._reaction.id],
        )
        if fva is not None:
            self._flux = self._flux.join(fva)

    def _string_flux(
        self,
        threshold: float,
        float_format: str,
    ) -> str:
        """
        Transform a flux data frame to a string.

        Parameters
        ----------
        threshold : float
            Hide fluxes below the threshold from being displayed.
        float_format : str
            Format string for floats.

        Returns
        -------
        str
            A string representation of the flux (with ranges).

        """

        if "minimum" in self._flux.columns and "maximum" in self._flux.columns:
            frame = self._flux.loc[
                (self._flux["flux"].abs() >= threshold)
                | (self._flux["minimum"].abs() >= threshold)
                | (self._flux["maximum"].abs() >= threshold),
                :,
            ].copy()
            return (
                f"{frame.at[self._reaction.id, 'flux']:{float_format}} "
                f"[{frame.at[self._reaction.id, 'minimum']:{float_format}}; "
                f"{frame.at[self._reaction.id, 'maximum']:{float_format}}]"
            )
        else:
            frame = self._flux.loc[self._flux["flux"].abs() >= threshold, :].copy()
            return f"{frame.at[self._reaction.id, 'flux']:{float_format}}"

    def to_string(
        self,
        names: bool = False,
        threshold: Optional[float] = None,
        float_format: str = ".4G",
        column_width: int = 79,
    ) -> str:
        """
        Return a pretty string representation of the reaction summary.

        Parameters
        ----------
        names : bool, optional
            Whether or not elements should be displayed by their common names
            (default False).
        threshold : float, optional
            Hide fluxes below the threshold from being displayed. If no value is
            given, the model tolerance is used (default None).
        float_format : str, optional
            Format string for floats (default '.4G').
        column_width : int, optional
            The maximum column width for each row (default 79).

        Returns
        -------
        str
            The summary formatted as a pretty string.

        """
        threshold = self._normalize_threshold(threshold)

        if names:
            header = shorten(self._reaction.name, width=column_width, placeholder="...")
        else:
            header = shorten(self._reaction.id, width=column_width, placeholder="...")

        flux = self._string_flux(threshold, float_format)

        return dedent(
            f"""
            {header}
            {'=' * len(header)}
            {self._reaction.build_reaction_string(use_metabolite_names=names)}
            Bounds: {self._reaction.lower_bound}, {self._reaction.upper_bound}
            Flux: {flux}
            """
        )

    def to_html(
        self,
        names: bool = False,
        threshold: Optional[float] = None,
        float_format: str = ".4G",
    ) -> str:
        """
        Return a rich HTML representation of the reaction summary.

        Parameters
        ----------
        names : bool, optional
            Whether or not elements should be displayed by their common names
            (default False).
        threshold : float, optional
            Hide fluxes below the threshold from being displayed. If no value is
            given, the model tolerance is used (default None).
        float_format : str, optional
            Format string for floats (default '.4G').

        Returns
        -------
        str
            The summary formatted as HTML.

        """
        threshold = self._normalize_threshold(threshold)

        if names:
            header = self._reaction.name
        else:
            header = self._reaction.id

        flux = self._string_flux(threshold, float_format)

        return (
            f"<h3>{header}</h3>"
            f"<p>{self._reaction.build_reaction_string(use_metabolite_names=names)}</p>"
            f"<p>Bounds: {self._reaction.lower_bound}, {self._reaction.upper_bound}</p>"
            f"<p>Flux: {flux}</p>"
        )



"""Helper functions for array operations and sampling."""

from typing import TYPE_CHECKING, NamedTuple, Optional, Union

import numpy as np
import pandas as pd


# Used to avoid cyclic reference and enable third-party static type checkers to work
if TYPE_CHECKING:
    from cobra import Model


try:
    from scipy.sparse import dok_matrix, lil_matrix
except ImportError:
    dok_matrix, lil_matrix = None, None


def create_stoichiometric_matrix(
    model: "Model", array_type: str = "dense", dtype: Optional[np.dtype] = None
) -> Union[np.ndarray, dok_matrix, lil_matrix, pd.DataFrame]:
    """Return a stoichiometric array representation of the given model.

    The the columns represent the reactions and rows represent
    metabolites. S[i,j] therefore contains the quantity of metabolite `i`
    produced (negative for consumed) by reaction `j`.

    Parameters
    ----------
    model : cobra.Model
        The cobra model to construct the matrix for.
    array_type : {"dense", "dok", "lil", "DataFrame"}
        The type of array to construct. "dense" will return a standard
        numpy.ndarray. "dok", or "lil" will construct a sparse array using
        scipy of the corresponding type. "DataFrame" will give a
        pandas.DataFrame with metabolite as indices and reaction as
        columns.
    dtype : numpy.dtype, optional
        The desired numpy data type for the array (default numpy.float64).

    Returns
    -------
    matrix of class `dtype`
        The stoichiometric matrix for the given model.

    Raises
    ------
    ValueError
        If sparse matrix is used and scipy is not installed.

    .. deprecated:: 0.18.1
              "DataFrame" option for `array_type` will be replaced with
              "frame" in future versions.

    """
    if array_type not in ("DataFrame", "dense") and not dok_matrix:
        raise ValueError("Sparse matrices require scipy.")

    if dtype is None:
        dtype = np.float64

    array_constructor = {
        "dense": np.zeros,
        "dok": dok_matrix,
        "lil": lil_matrix,
        "DataFrame": np.zeros,
    }

    n_metabolites = len(model.metabolites)
    n_reactions = len(model.reactions)
    array = array_constructor[array_type]((n_metabolites, n_reactions), dtype=dtype)

    m_ind = model.metabolites.index
    r_ind = model.reactions.index

    for reaction in model.reactions:
        for metabolite, stoich in reaction.metabolites.items():
            array[m_ind(metabolite), r_ind(reaction)] = stoich

    if array_type == "DataFrame":
        metabolite_ids = [met.id for met in model.metabolites]
        reaction_ids = [rxn.id for rxn in model.reactions]
        return pd.DataFrame(array, index=metabolite_ids, columns=reaction_ids)

    else:
        return array


def nullspace(A: np.ndarray, atol: float = 1e-13, rtol: float = 0.0) -> np.ndarray:
    r"""Compute an approximate basis for the nullspace of A.

    The algorithm used by this function is based on the Singular Value
    Decomposition (SVD) of `A`.

    Parameters
    ----------
    A : numpy.ndarray
        `A` should be at most 2-D. 1-D array with length k will be treated
        as a 2-D with shape (1, k).
    atol : float, optional
        The absolute tolerance for a zero singular value. Singular values
        smaller than `atol` are considered to be zero (default 1e-13).
    rtol : float, optional
        The relative tolerance. Singular values less than `rtol * smax` are
        considered to be zero, where `smax` is the largest singular value
        (default 0.0).

    Returns
    -------
    numpy.ndarray
        If `A` is an array with shape (m, k), then `ns` will be an array
        with shape (k, n), where `n` is the estimated dimension of the
        nullspace of `A`.  The columns of `ns` are a basis for the
        nullspace; each element in numpy.dot(A, ns) will be approximately
        zero.

    Notes
    -----
    This is taken from the numpy cookbook.

    If both `atol` and `rtol` are positive, the combined tolerance is the
    maximum of the two; that is:

    .. math:: \mathtt{tol} = \max(\mathtt{atol}, \mathtt{rtol} * \mathtt{smax})

    Singular values smaller than `tol` are considered to be zero.

    """
    A = np.atleast_2d(A)
    _, s, vh = np.linalg.svd(A)
    tol = max(atol, rtol * s[0])
    nnz = (s >= tol).sum()
    ns = vh[nnz:].conj().T
    return ns


def constraint_matrices(
    model: "Model",
    array_type: str = "dense",
    zero_tol: float = 1e-6,
) -> NamedTuple:
    """Create a matrix representation of the problem.

    This is used for alternative solution approaches that do not use
    "optlang". The function will construct the equality matrix,
    inequality matrix and bounds for the complete problem.

    Parameters
    ----------
    model : cobra.Model
        The model from which to obtain the LP problem.
    array_type : {"dense", "dok", "lil", "DataFrame"}
        The type of array to construct. "dense" will return a standard
        numpy.ndarray. "dok", or "lil" will construct a sparse array using
        scipy of the corresponding type. "DataFrame" will give a
        pandas.DataFrame with metabolite as indices and reaction as
        columns.
    zero_tol : float, optional
        The zero tolerance used to judge whether two bounds are the same
        (default 1e-6).

    Returns
    -------
    NamedTuple
        A named tuple consisting of 6 matrices and 2 vectors:
        - "equalities" is a matrix `S` such that `S * vars = b`. It
          includes a row for each constraint and one column for each
          variable.
        - "b" is the right side of the equality equation such that
          `S * vars = b`.
        - "inequalities" is a matrix M such that `lb <= M * vars <= ub`.
          It contains a row for each inequality and as many columns as
          variables.
        - "bounds" is a compound matrix [lb ub] containing the lower and
          upper bounds for the inequality constraints in M.
        - "variable_fixed" is a boolean vector indicating whether the
          variable at that index is fixed (`lower bound == upper_bound`)
          and is thus bounded by an equality constraint.
        - "variable_bounds" is a compound matrix `[lb ub]` containing the
          lower and upper bounds for all variables.

    Notes
    -----
    To accomodate non-zero equalities, the problem will add the variable
    "const_one" which is a variable that equals one.

    .. deprecated:: 0.18.1
              "DataFrame" option for `array_type` will be replaced with
              "frame" in future versions.

    """
    if array_type not in ("DataFrame", "dense") and not dok_matrix:
        raise ValueError("Sparse matrices require scipy.")

    array_builder = {
        "dense": np.array,
        "dok": dok_matrix,
        "lil": lil_matrix,
        "DataFrame": pd.DataFrame,
    }[array_type]

    Problem = NamedTuple(
        "Problem",
        [
            ("equalities", Union[np.ndarray, dok_matrix, lil_matrix, pd.DataFrame]),
            ("b", np.ndarray),
            ("inequalities", Union[np.ndarray, dok_matrix, lil_matrix, pd.DataFrame]),
            ("bounds", Union[np.ndarray, dok_matrix, lil_matrix, pd.DataFrame]),
            ("variable_fixed", np.ndarray),
            (
                "variable_bounds",
                Union[np.ndarray, dok_matrix, lil_matrix, pd.DataFrame],
            ),
        ],
    )
    equality_rows = []
    inequality_rows = []
    inequality_bounds = []
    b = []

    for const in model.constraints:
        lb = -np.inf if const.lb is None else const.lb
        ub = np.inf if const.ub is None else const.ub
        equality = (ub - lb) < zero_tol
        coefs = const.get_linear_coefficients(model.variables)
        coefs = [coefs[v] for v in model.variables]
        if equality:
            b.append(lb if abs(lb) > zero_tol else 0.0)
            equality_rows.append(coefs)
        else:
            inequality_rows.append(coefs)
            inequality_bounds.append([lb, ub])

    var_bounds = np.array([[v.lb, v.ub] for v in model.variables])
    fixed = var_bounds[:, 1] - var_bounds[:, 0] < zero_tol

    results = Problem(
        equalities=array_builder(equality_rows),
        b=np.array(b),
        inequalities=array_builder(inequality_rows),
        bounds=array_builder(inequality_bounds),
        variable_fixed=np.array(fixed),
        variable_bounds=array_builder(var_bounds),
    )

    return results



"""Provide parsimonious FBA implementation."""

from itertools import chain
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union
from warnings import warn

from optlang.symbolics import Zero

from ..core.solution import get_solution
from ..util import solver as sutil


if TYPE_CHECKING:
    from optlang.interface import Objective

    from cobra import Model, Reaction, Solution


def optimize_minimal_flux(
    *args, **kwargs
) -> Callable[["Model", float, Union[Dict, "Objective"], List["Reaction"]], "Solution"]:
    """Perform basic pFBA to minimize total flux.

    .. deprecated:: 0.6.0a4
            `optimize_minimal_flux` will be removed in cobrapy 1.0.0, it is
            replaced by `pfba`.

    Parameters
    ----------
    *args: Any
        Non-keyword variable-length arguments.
    **kwargs: Any
        Keyword-only variable-length arguments.

    Returns
    -------
    A function performing the parsimonious FBA.

    """
    warn("optimize_minimal_flux has been renamed to pfba", DeprecationWarning)
    return pfba(*args, **kwargs)


def pfba(
    model: "Model",
    fraction_of_optimum: float = 1.0,
    objective: Union[Dict, "Objective", None] = None,
    reactions: Optional[List["Reaction"]] = None,
) -> "Solution":
    """Perform basic pFBA (parsimonious Enzyme Usage Flux Balance Analysis).

    pFBA [1] adds the minimization of all fluxes the the objective of the
    model. This approach is motivated by the idea that high fluxes have a
    higher enzyme turn-over and that since producing enzymes is costly,
    the cell will try to minimize overall flux while still maximizing the
    original objective function, e.g. the growth rate.

    Parameters
    ----------
    model : cobra.Model
        The model to perform pFBA on.
    fraction_of_optimum : float, optional
        The fraction of optimum which must be maintained. The original
        objective reaction is constrained to be greater than maximal value
        times the `fraction_of_optimum` (default 1.0).
    objective : dict or cobra.Model.objective, optional
        A desired objective to use during optimization in addition to the
        pFBA objective. Dictionaries (reaction as key, coefficient as value)
        can be used for linear objectives (default None).
    reactions : list of cobra.Reaction, optional
        List of cobra.Reaction. Implies `return_frame` to be true. Only
        return fluxes for the given reactions. Faster than fetching all
        fluxes if only a few are needed (default None).

    Returns
    -------
    cobra.Solution
        The solution object to the optimized model with pFBA constraints
        added.

    References
    ----------
    .. [1] Lewis, N. E., Hixson, K. K., Conrad, T. M., Lerman, J. A.,
       Charusanti, P., Polpitiya, A. D., Palsson, B. O. (2010). Omic data
       from evolved E. coli are consistent with computed optimal growth from
       genome-scale models. Molecular Systems Biology, 6,
       390. doi:10.1038/msb.2010.47

    """
    reactions = (
        model.reactions if reactions is None else model.reactions.get_by_any(reactions)
    )
    with model as m:
        add_pfba(m, objective=objective, fraction_of_optimum=fraction_of_optimum)
        m.slim_optimize(error_value=None)
        solution = get_solution(m, reactions=reactions)
    return solution


def add_pfba(
    model: "Model",
    objective: Union[Dict, "Objective", None] = None,
    fraction_of_optimum: float = 1.0,
) -> None:
    """Add pFBA objective to the `model`.

    This adds objective to minimize the summed flux of all reactions to the
    current objective.

    Parameters
    ----------
    model : cobra.Model
        The model to add the objective to.
    objective : dict or cobra.Model.objective, optional
        A desired objective to use during optimization in addition to the
        pFBA objective. Dictionaries (reaction as key, coefficient as value)
        can be used for linear objectives (default None).
    fraction_of_optimum : float, optional
        Fraction of optimum which must be maintained. The original objective
        reaction is constrained to be greater than maximal value times the
        `fraction_of_optimum`.

    See Also
    -------
    pfba

    """
    if objective is not None:
        model.objective = objective
    if model.solver.objective.name == "_pfba_objective":
        raise ValueError("The model already has a pFBA objective.")
    sutil.fix_objective_as_constraint(model, fraction=fraction_of_optimum)
    reaction_variables = (
        (rxn.forward_variable, rxn.reverse_variable) for rxn in model.reactions
    )
    variables = chain(*reaction_variables)
    model.objective = model.problem.Objective(
        Zero, direction="min", sloppy=True, name="_pfba_objective"
    )
    model.objective.set_linear_coefficients({v: 1.0 for v in variables})



#!/usr/bin/env python


"""
Regenerate model pickle files.

This should be performed after updating core classes in order to prevent suble
bugs.

"""


from collections import OrderedDict
from json import dump as json_dump
from pickle import dump, load

import cobra
from cobra.io import (
    load_matlab_model,
    read_sbml_model,
    save_json_model,
    save_matlab_model,
    save_yaml_model,
    write_sbml_model,
)


config = cobra.Configuration()
config.solver = "glpk"


if __name__ == "__main__":
    # ecoli
    ecoli_model = read_sbml_model("iJO1366.xml.gz")
    with open("iJO1366.pickle", "wb") as outfile:
        dump(ecoli_model, outfile, protocol=2)

    # salmonella
    salmonella = read_sbml_model("salmonella.xml")
    with open("salmonella.genes", "rb") as infile:
        gene_names = load(infile)
    for gene in salmonella.genes:
        gene.name = gene_names[gene.id]
    with open("salmonella.media", "rb") as infile:
        salmonella.media_compositions = load(infile)
    with open("salmonella.pickle", "wb") as outfile:
        dump(salmonella, outfile, protocol=2)

    # create mini model from textbook
    textbook = read_sbml_model("textbook.xml.gz")
    mini = cobra.Model("mini_textbook")
    mini.compartments = textbook.compartments

    for r in textbook.reactions:
        if r.id in (
            "GLCpts",
            "PGI",
            "PFK",
            "FBA",
            "TPI",
            "GAPD",
            "PGK",
            "PGM",
            "ENO",
            "PYK",
            "EX_glc__D_e",
            "EX_h_e",
            "H2Ot",
            "ATPM",
            "PIt2r",
        ):
            mini.add_reaction(r.copy())
    mini.reactions.ATPM.upper_bound = mini.reactions.PGI.upper_bound
    mini.objective = ["PFK", "ATPM"]  # No biomass, 2 reactions

    # add in some information from iJO1366
    mini.add_reaction(ecoli_model.reactions.LDH_D.copy())
    mini.add_reaction(ecoli_model.reactions.EX_lac__D_e.copy())
    r = cobra.Reaction("D_LACt2")
    mini.add_reaction(r)
    mini.reactions.GLCpts.gene_reaction_rule = (
        ecoli_model.reactions.GLCptspp.gene_reaction_rule
    )

    # adjust bounds
    for i in ["ATPM", "D_LACt2", "EX_lac__D_e", "LDH_D"]:
        mini.reactions.get_by_id(i).upper_bound = mini.reactions.PGI.upper_bound
    for i in ["D_LACt2", "LDH_D"]:
        mini.reactions.get_by_id(i).lower_bound = mini.reactions.PGI.lower_bound
    # set names and annotation
    for g in mini.genes:
        try:
            tg = textbook.genes.get_by_id(g.id)
        except KeyError:
            continue
        g.name = tg.name
        g.annotation = tg.annotation
    mini.reactions.sort()
    mini.genes.sort()
    mini.metabolites.sort()
    # output to various formats
    with open("mini.pickle", "wb") as outfile:
        dump(mini, outfile, protocol=2)
    save_matlab_model(mini, "mini.mat")
    save_json_model(mini, "mini.json", pretty=True)
    save_yaml_model(mini, "mini.yml")
    write_sbml_model(mini, "mini_fbc2.xml")
    write_sbml_model(mini, "mini_fbc2.xml.bz2")
    write_sbml_model(mini, "mini_fbc2.xml.gz")
    write_sbml_model(mini, "mini_cobra.xml")
    raven = load_matlab_model("raven.mat")
    with open("raven.pickle", "wb") as outfile:
        dump(raven, outfile, protocol=2)

    # TODO:these need a reference solutions rather than circular solution checking!

    # fva results
    fva_result = cobra.flux_analysis.flux_variability_analysis(textbook)
    clean_result = OrderedDict()
    for key in sorted(fva_result):
        clean_result[key] = {k: round(v, 5) for k, v in fva_result[key].items()}
    with open("textbook_fva.json", "w") as outfile:
        json_dump(clean_result, outfile)

    # fva with pfba constraint
    fva_result = cobra.flux_analysis.flux_variability_analysis(
        textbook, pfba_factor=1.1
    )
    clean_result = OrderedDict()
    for key in sorted(fva_result):
        clean_result[key] = {k: round(v, 5) for k, v in fva_result[key].items()}
    with open("textbook_pfba_fva.json", "w") as outfile:
        json_dump(clean_result, outfile)

    # textbook solution
    solution = cobra.flux_analysis.parsimonious.pfba(textbook)
    with open("textbook_solution.pickle", "wb") as f:
        dump(solution, f, protocol=2)



"""Test functionalities of model component annotation functions."""

from cobra.core import Model, Reaction
from cobra.manipulation import add_SBO


def test_sbo_annotation(model: Model) -> None:
    """Test SBO annotation function."""
    rxns = model.reactions
    rxns.EX_o2_e.annotation.clear()
    fake_DM = Reaction("DM_h_c")
    model.add_reactions([fake_DM])
    fake_DM.add_metabolites({model.metabolites.get_by_id("h_c"): -1})
    # this exchange will be set wrong. The function should not overwrite
    # an existing SBO annotation
    rxns.get_by_id("EX_h_e").annotation["sbo"] = "SBO:0000628"
    add_SBO(model)
    assert rxns.EX_o2_e.annotation["sbo"] == "SBO:0000627"
    assert rxns.DM_h_c.annotation["sbo"] == "SBO:0000628"
    assert rxns.EX_h_e.annotation["sbo"] == "SBO:0000628"



"""Provide function for annotating demand and exchange reactions."""

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from cobra import Model


def add_SBO(model: "Model") -> None:
    """Add SBO terms for demands and exchanges.

    This works for models which follow the standard convention for
    constructing and naming these reactions.

    The reaction should only contain the single metabolite being exchanged,
    and the id should be EX_<met_id> or DM_<met_id> .

    Parameters
    ----------
    model: cobra.Model
        The model whose demand and exchange reactions need to be annotated.

    """
    for r in model.reactions:
        # don't annotate already annotated reactions
        if r.annotation.get("sbo"):
            continue
        # only doing exchanges
        if len(r.metabolites) != 1:
            continue
        met_id = list(r._metabolites)[0].id
        if r.id.startswith("EX_") and r.id == "EX_" + met_id:
            r.annotation["sbo"] = "SBO:0000627"
        elif r.id.startswith("DM_") and r.id == "DM_" + met_id:
            r.annotation["sbo"] = "SBO:0000628"



"""Provide lists and annotations for compartment names and reactions.

Please send a PR if you want to add something here :)

"""

# A dictionary having keys as reaction types and keys as prefixes of
# reaction IDs that usually indicate that the reaction is *not* a reaction
# of the specified type.
excludes = {
    "demand": ["SN_", "SK_", "sink", "EX_", "exchange"],
    "exchange": [
        "demand",
        "DM_",
        "biosynthesis",
        "transcription",
        "replication",
        "SN_",
        "SK_",
        "sink",
    ],
    "sink": [
        "demand",
        "DM_",
        "biosynthesis",
        "transcription",
        "replication",
        "EX_",
        "exchange",
    ],
}


# A dictionary having SBO term identifiers as values and boundary types
# as keys.
sbo_terms = {
    "demand": "SBO:0000628",
    "exchange": "SBO:0000627",
    "sink": "SBO:0000632",
    "biomass": "SBO:0000629",
    "pseudoreaction": "SBO:0000631",
}


# A dictionary having keys as common compartment abbreviations and values
# as alternative names.
compartment_shortlist = {
    "ce": ["cell envelope"],
    "c": [
        "cytoplasm",
        "cytosol",
        "default",
        "in",
        "intra cellular",
        "intracellular",
        "intracellular region",
        "intracellular space",
    ],
    "er": ["endoplasmic reticulum"],
    "erm": ["endoplasmic reticulum membrane"],
    "e": [
        "extracellular",
        "extraorganism",
        "out",
        "extracellular space",
        "extra organism",
        "extra cellular",
        "extra-organism",
        "external",
        "external medium",
    ],
    "f": ["flagellum", "bacterial-type flagellum"],
    "g": ["golgi", "golgi apparatus"],
    "gm": ["golgi membrane"],
    "h": ["chloroplast"],
    "l": ["lysosome"],
    "im": ["mitochondrial intermembrane space"],
    "mm": ["mitochondrial membrane"],
    "m": ["mitochondrion", "mitochondria"],
    "n": ["nucleus"],
    "p": ["periplasm", "periplasmic space"],
    "x": ["peroxisome", "glyoxysome"],
    "u": ["thylakoid"],
    "vm": ["vacuolar membrane"],
    "v": ["vacuole"],
    "w": ["cell wall"],
    "s": ["eyespot", "eyespot apparatus", "stigma"],
}



# -*- coding: utf-8 -*-

"""Define the Metabolite class."""

from __future__ import absolute_import

import re
from warnings import warn

from future.utils import raise_from, raise_with_traceback

from cobra.core.formula import elements_and_molecular_weights
from cobra.core.species import Species
from cobra.exceptions import OptimizationError
from cobra.util.solver import check_solver_status
from cobra.util.util import format_long_string


# Numbers are not required because of the |(?=[A-Z])? block. See the
# discussion in https://github.com/opencobra/cobrapy/issues/128 for
# more details.
element_re = re.compile("([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)")


class Metabolite(Species):
    """Metabolite is a class for holding information regarding
    a metabolite in a cobra.Reaction object.

    Parameters
    ----------
    id : str
        the identifier to associate with the metabolite
    formula : str
        Chemical formula (e.g. H2O)
    name : str
        A human readable name.
    charge : float
       The charge number of the metabolite
    compartment: str or None
       Compartment of the metabolite.
    """

    def __init__(self, id=None, formula=None, name="", charge=None, compartment=None):
        Species.__init__(self, id, name)
        self.formula = formula
        # because in a Model a metabolite may participate in multiple Reactions
        self.compartment = compartment
        self.charge = charge

        self._bound = 0.0

    def _set_id_with_model(self, value):
        if value in self.model.metabolites:
            raise ValueError(
                "The model already contains a metabolite with " "the id:", value
            )
        self.model.constraints[self.id].name = value
        self._id = value
        self.model.metabolites._generate_index()

    @property
    def constraint(self):
        """Get the constraints associated with this metabolite from the solve

        Returns
        -------
        optlang.<interface>.Constraint
            the optlang constraint for this metabolite
        """
        if self.model is not None:
            return self.model.constraints[self.id]

    @property
    def elements(self):
        """Dictionary of elements as keys and their count in the metabolite
        as integer. When set, the `formula` property is update accordingly"""
        tmp_formula = self.formula
        if tmp_formula is None:
            return {}
        # necessary for some old pickles which use the deprecated
        # Formula class
        tmp_formula = str(self.formula)
        # commonly occurring characters in incorrectly constructed formulas
        if "*" in tmp_formula:
            warn("invalid character '*' found in formula '%s'" % self.formula)
            tmp_formula = tmp_formula.replace("*", "")
        if "(" in tmp_formula or ")" in tmp_formula:
            warn("invalid formula (has parenthesis) in '%s'" % self.formula)
            return None
        composition = {}
        parsed = element_re.findall(tmp_formula)
        for (element, count) in parsed:
            if count == "":
                count = 1
            else:
                try:
                    count = float(count)
                    int_count = int(count)
                    if count == int_count:
                        count = int_count
                    else:
                        warn(
                            "%s is not an integer (in formula %s)"
                            % (count, self.formula)
                        )
                except ValueError:
                    warn("failed to parse %s (in formula %s)" % (count, self.formula))
                    return None
            if element in composition:
                composition[element] += count
            else:
                composition[element] = count
        return composition

    @elements.setter
    def elements(self, elements_dict):
        def stringify(element, number):
            return element if number == 1 else element + str(number)

        self.formula = "".join(
            stringify(e, n) for e, n in sorted(elements_dict.items())
        )

    @property
    def formula_weight(self):
        """Calculate the formula weight"""
        try:
            return sum(
                [
                    count * elements_and_molecular_weights[element]
                    for element, count in self.elements.items()
                ]
            )
        except KeyError as e:
            warn("The element %s does not appear in the periodic table" % e)

    @property
    def y(self):
        """The shadow price for the metabolite in the most recent solution

        Shadow prices are computed from the dual values of the bounds in
        the solution.

        """
        warn("Please use metabolite.shadow_price instead.", DeprecationWarning)
        return self.shadow_price

    @property
    def shadow_price(self):
        """
        The shadow price in the most recent solution.

        Shadow price is the dual value of the corresponding constraint in the
        model.

        Warnings
        --------
        * Accessing shadow prices through a `Solution` object is the safer,
          preferred, and only guaranteed to be correct way. You can see how to
          do so easily in the examples.
        * Shadow price is retrieved from the currently defined
          `self._model.solver`. The solver status is checked but there are no
          guarantees that the current solver state is the one you are looking
          for.
        * If you modify the underlying model after an optimization, you will
          retrieve the old optimization values.

        Raises
        ------
        RuntimeError
            If the underlying model was never optimized beforehand or the
            metabolite is not part of a model.
        OptimizationError
            If the solver status is anything other than 'optimal'.

        Examples
        --------
        >>> import cobra
        >>> import cobra.test
        >>> model = cobra.test.create_test_model("textbook")
        >>> solution = model.optimize()
        >>> model.metabolites.glc__D_e.shadow_price
        -0.09166474637510488
        >>> solution.shadow_prices.glc__D_e
        -0.091664746375104883
        """
        try:
            check_solver_status(self._model.solver.status)
            return self._model.constraints[self.id].dual
        except AttributeError:
            raise RuntimeError("metabolite '{}' is not part of a model".format(self.id))
        # Due to below all-catch, which sucks, need to reraise these.
        except (RuntimeError, OptimizationError) as err:
            raise_with_traceback(err)
        # Would love to catch CplexSolverError and GurobiError here.
        except Exception as err:
            raise_from(
                OptimizationError(
                    "Likely no solution exists. Original solver message: {}."
                    "".format(str(err))
                ),
                err,
            )

    def remove_from_model(self, destructive=False):
        """Removes the association from self.model

        The change is reverted upon exit when using the model as a context.

        Parameters
        ----------
        destructive : bool
            If False then the metabolite is removed from all
            associated reactions.  If True then all associated
            reactions are removed from the Model.
        """
        self._model.remove_metabolites(self, destructive)

    def summary(self, solution=None, fva=None):
        """
        Create a summary of the producing and consuming fluxes.

        Parameters
        ----------
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution (default None).
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched (default None).

        Returns
        -------
        cobra.summary.MetaboliteSummary

        See Also
        --------
        Reaction.summary
        Model.summary

        """
        from cobra.summary import MetaboliteSummary

        return MetaboliteSummary(
            metabolite=self,
            model=self._model,
            solution=solution,
            fva=fva,
        )

    def _repr_html_(self):
        return """
        <table>
            <tr>
                <td><strong>Metabolite identifier</strong></td><td>{id}</td>
            </tr><tr>
                <td><strong>Name</strong></td><td>{name}</td>
            </tr><tr>
                <td><strong>Memory address</strong></td>
                <td>{address}</td>
            </tr><tr>
                <td><strong>Formula</strong></td><td>{formula}</td>
            </tr><tr>
                <td><strong>Compartment</strong></td><td>{compartment}</td>
            </tr><tr>
                <td><strong>In {n_reactions} reaction(s)</strong></td><td>
                    {reactions}</td>
            </tr>
        </table>""".format(
            id=self.id,
            name=format_long_string(self.name),
            formula=self.formula,
            address="0x0%x" % id(self),
            compartment=self.compartment,
            n_reactions=len(self.reactions),
            reactions=format_long_string(", ".join(r.id for r in self.reactions), 200),
        )



"""Provide functions to modify model components."""

from ast import NodeTransformer
from itertools import chain
from typing import TYPE_CHECKING, Dict

from ..core.gene import ast2str
from .delete import get_compiled_gene_reaction_rules


if TYPE_CHECKING:
    from cobra import Gene, Model

# Set of tuples of operators and their corresponding textual form
_renames = (
    (".", "_DOT_"),
    ("(", "_LPAREN_"),
    (")", "_RPAREN_"),
    ("-", "__"),
    ("[", "_LSQBKT"),
    ("]", "_RSQBKT"),
    (",", "_COMMA_"),
    (":", "_COLON_"),
    (">", "_GT_"),
    ("<", "_LT"),
    ("/", "_FLASH"),
    ("\\", "_BSLASH"),
    ("+", "_PLUS_"),
    ("=", "_EQ_"),
    (" ", "_SPACE_"),
    ("'", "_SQUOT_"),
    ('"', "_DQUOT_"),
)


def _escape_str_id(id_str: str) -> str:
    """Make a single string ID SBML compliant.

    Parameters
    ----------
    id_str: str
        The ID string to operate on.

    Returns
    -------
    str
        The SBML compliant ID string.

    """
    for c in ("'", '"'):
        if id_str.startswith(c) and id_str.endswith(c) and id_str.count(c) == 2:
            id_str = id_str.strip(c)
    for char, escaped_char in _renames:
        id_str = id_str.replace(char, escaped_char)
    return id_str


class _GeneEscaper(NodeTransformer):
    """Class to represent a gene ID escaper."""

    def visit_Name(self, node: "Gene") -> "Gene":
        """Escape string ID.

        Parameters
        ----------
        node: cobra.Gene
            The gene object to work on.

        Returns
        -------
        cobra.Gene
            The gene object whose ID has been escaped.

        """
        node.id = _escape_str_id(node.id)
        return node


def escape_ID(model: "Model") -> None:
    """Make all model component object IDs SBML compliant.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.

    """
    for x in chain([model], model.metabolites, model.reactions, model.genes):
        x.id = _escape_str_id(x.id)
    model.repair()
    gene_renamer = _GeneEscaper()
    for rxn, rule in get_compiled_gene_reaction_rules(model).items():
        if rule is not None:
            rxn._gene_reaction_rule = ast2str(gene_renamer.visit(rule))


class _Renamer(NodeTransformer):
    """
    Class to represent a gene renamer.

    Parameters
    ----------
    rename_dict: dict of {str: str}
        The dictionary having keys as old gene names
        and value as new gene names.

    """

    def __init__(self, rename_dict: Dict[str, str], **kwargs) -> None:
        """Initialize a new object.

        Other Parameters
        ----------------
        **kwargs:
            Further keyword arguments are passed on to the parent class.

        """
        super().__init__(**kwargs)
        self.rename_dict = rename_dict

    def visit_Name(self, node: "Gene") -> "Gene":
        """Rename a gene.

        Parameters
        ----------
        node: cobra.Gene
            The gene to rename.

        Returns
        -------
        cobra.Gene
            The renamed gene object.

        """
        node.id = self.rename_dict.get(node.id, node.id)
        return node


def rename_genes(model: "Model", rename_dict: Dict[str, str]) -> None:
    """Rename genes in a model from the `rename_dict`.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.
    rename_dict: dict of {str: str}
        The dictionary having keys as old gene names
        and value as new gene names.

    """
    recompute_reactions = set()  # need to recompute related genes
    remove_genes = []
    for old_name, new_name in rename_dict.items():
        # undefined if there a value matches a different key
        try:
            gene_index = model.genes.index(old_name)
        except ValueError:
            gene_index = None
        old_gene_present = gene_index is not None
        new_gene_present = new_name in model.genes
        if old_gene_present and new_gene_present:
            old_gene = model.genes.get_by_id(old_name)
            # Added in case not renaming some genes:
            if old_gene is not model.genes.get_by_id(new_name):
                remove_genes.append(old_gene)
                recompute_reactions.update(old_gene._reaction)
        elif old_gene_present and not new_gene_present:
            # rename old gene to new gene
            gene = model.genes[gene_index]
            # trick DictList into updating index
            model.genes._dict.pop(gene.id)  # ugh
            gene.id = new_name
            model.genes[gene_index] = gene
        elif not old_gene_present and new_gene_present:
            pass
        else:  # if not old gene_present and not new_gene_present
            # the new gene's _model will be set by repair
            # This would add genes from rename_dict
            # that are not associated with a rxn
            # cobra_model.genes.append(Gene(new_name))
            pass
    model.repair()

    gene_renamer = _Renamer(rename_dict)
    for rxn, rule in get_compiled_gene_reaction_rules(model).items():
        if rule is not None:
            rxn._gene_reaction_rule = ast2str(gene_renamer.visit(rule))

    for rxn in recompute_reactions:
        rxn.gene_reaction_rule = rxn._gene_reaction_rule
    for i in remove_genes:
        model.genes.remove(i)



"""Test functionalities of model component pruning functions."""

from itertools import chain
from typing import List, Set

from cobra.core import Gene, Metabolite, Model, Reaction
from cobra.manipulation import (
    delete_model_genes,
    find_gene_knockout_reactions,
    get_compiled_gene_reaction_rules,
    prune_unused_metabolites,
    prune_unused_reactions,
    remove_genes,
    undelete_model_genes,
)


def test_prune_unused_metabolites_output_type(model: Model) -> None:
    """Test the output type of unused metabolites pruning."""
    [model.reactions.get_by_id(x).remove_from_model() for x in ["RPI", "RPE", "GND"]]
    model_pruned, unused = prune_unused_metabolites(model)
    assert isinstance(model_pruned, Model)
    # test that the output contains metabolite objects
    assert isinstance(unused[0], Metabolite)


def test_prune_unused_metabolites_sanity(model: Model) -> None:
    """Test the sanity of unused metabolites pruning."""
    metabolite1 = model.metabolites.ru5p__D_c
    metabolite2 = model.metabolites.akg_e
    metabolite3 = model.metabolites.akg_c
    reactions = set(
        chain(metabolite1.reactions, metabolite2.reactions, metabolite3.reactions)
    )
    model.remove_reactions(reactions)
    model_pruned, unused = prune_unused_metabolites(model)
    assert metabolite1 in model.metabolites
    assert metabolite2 in model.metabolites
    assert metabolite3 in model.metabolites
    # test that the unused metabolites are not used in the model
    assert metabolite1 not in model_pruned.metabolites
    assert metabolite2 not in model_pruned.metabolites
    assert metabolite3 not in model_pruned.metabolites


def test_prune_unused_reactions_output_type(model: Model) -> None:
    """Test the output type of unused reactions pruning."""
    reaction = Reaction("foo")
    model.add_reactions([reaction])
    model_pruned, unused = prune_unused_reactions(model)
    assert isinstance(model_pruned, Model)
    # test that the output contains reaction objects
    assert isinstance(unused[0], Reaction)


def test_prune_unused_rxns_functionality(model: Model) -> None:
    """Test the sanity of unused reactions pruning."""
    for x in ["foo1", "foo2", "foo3"]:
        model.add_reactions([Reaction(x)])

    model_pruned, unused = prune_unused_reactions(model)
    assert "foo1" in model.reactions
    assert "foo2" in model.reactions
    assert "foo3" in model.reactions
    # test that the unused reactions are not used in the model
    assert "foo1" not in model_pruned.reactions
    assert "foo2" not in model_pruned.reactions
    assert "foo3" not in model_pruned.reactions


def _find_gene_knockout_reactions_fast(
    m: Model, gene_list: List[Gene]
) -> List[Reaction]:
    """Quickly find gene knockout reactions."""
    compiled_rules = get_compiled_gene_reaction_rules(m)
    return find_gene_knockout_reactions(
        m, gene_list, compiled_gene_reaction_rules=compiled_rules
    )


def _get_removed(m: Model) -> Set[str]:
    """Get trimmed reactions."""
    return {x.id for x in m._trimmed_reactions}


def _gene_knockout_computation(
    m: Model, gene_ids: List[str], expected_reaction_ids: List[str]
) -> None:
    """Compute gene knockout."""
    genes = [m.genes.get_by_id(i) for i in gene_ids]
    expected_reactions = {m.reactions.get_by_id(i) for i in expected_reaction_ids}
    removed1 = set(find_gene_knockout_reactions(m, genes))
    removed2 = set(_find_gene_knockout_reactions_fast(m, genes))
    assert removed1 == expected_reactions
    assert removed2 == expected_reactions
    delete_model_genes(m, gene_ids, cumulative_deletions=False)
    assert _get_removed(m) == expected_reaction_ids
    undelete_model_genes(m)


def test_gene_knockout(salmonella: Model) -> None:
    """Test gene knockout."""
    gene_list = ["STM1067", "STM0227"]
    dependent_reactions = {
        "3HAD121",
        "3HAD160",
        "3HAD80",
        "3HAD140",
        "3HAD180",
        "3HAD100",
        "3HAD181",
        "3HAD120",
        "3HAD60",
        "3HAD141",
        "3HAD161",
        "T2DECAI",
        "3HAD40",
    }
    _gene_knockout_computation(salmonella, gene_list, dependent_reactions)
    _gene_knockout_computation(salmonella, ["STM4221"], {"PGI"})
    _gene_knockout_computation(salmonella, ["STM1746.S"], {"4PEPTabcpp"})
    # test cumulative behavior
    delete_model_genes(salmonella, gene_list[:1])
    delete_model_genes(salmonella, gene_list[1:], cumulative_deletions=True)
    delete_model_genes(salmonella, ["STM4221"], cumulative_deletions=True)
    dependent_reactions.add("PGI")
    assert _get_removed(salmonella) == dependent_reactions
    # non-cumulative following cumulative
    delete_model_genes(salmonella, ["STM4221"], cumulative_deletions=False)
    assert _get_removed(salmonella) == {"PGI"}
    # make sure on reset that the bounds are correct
    reset_bound = salmonella.reactions.get_by_id("T2DECAI").upper_bound
    assert reset_bound == 1000.0
    # test computation when gene name is a subset of another
    test_model = Model()
    test_reaction_1 = Reaction("test1")
    test_reaction_1.gene_reaction_rule = "eggs or (spam and eggspam)"
    test_model.add_reactions([test_reaction_1])
    _gene_knockout_computation(test_model, ["eggs"], set())
    _gene_knockout_computation(test_model, ["eggs", "spam"], {"test1"})
    # test computation with nested boolean expression
    test_reaction_1.gene_reaction_rule = "g1 and g2 and (g3 or g4 or (g5 and g6))"
    _gene_knockout_computation(test_model, ["g3"], set())
    _gene_knockout_computation(test_model, ["g1"], {"test1"})
    _gene_knockout_computation(test_model, ["g5"], set())
    _gene_knockout_computation(test_model, ["g3", "g4", "g5"], {"test1"})
    # test computation when gene names are python expressions
    test_reaction_1.gene_reaction_rule = "g1 and (for or in)"
    _gene_knockout_computation(test_model, ["for", "in"], {"test1"})
    _gene_knockout_computation(test_model, ["for"], set())
    test_reaction_1.gene_reaction_rule = "g1 and g2 and g2.conjugate"
    _gene_knockout_computation(test_model, ["g2"], {"test1"})
    _gene_knockout_computation(test_model, ["g2.conjugate"], {"test1"})
    test_reaction_1.gene_reaction_rule = "g1 and (try:' or 'except:1)"
    _gene_knockout_computation(test_model, ["try:'"], set())
    _gene_knockout_computation(test_model, ["try:'", "'except:1"], {"test1"})


def test_remove_genes() -> None:
    """Test gene removal."""
    m = Model("test")
    m.add_reactions([Reaction("r" + str(i + 1)) for i in range(8)])
    assert len(m.reactions) == 8
    rxns = m.reactions
    rxns.r1.gene_reaction_rule = "(a and b) or (c and a)"
    rxns.r2.gene_reaction_rule = "(a and b and d and e)"
    rxns.r3.gene_reaction_rule = "(a and b) or (b and c)"
    rxns.r4.gene_reaction_rule = "(f and b) or (b and c)"
    rxns.r5.gene_reaction_rule = "x"
    rxns.r6.gene_reaction_rule = "y"
    rxns.r7.gene_reaction_rule = "x or     z"
    rxns.r8.gene_reaction_rule = ""
    assert "a" in m.genes
    assert "x" in m.genes
    remove_genes(m, ["a"], remove_reactions=False)
    assert "a" not in m.genes
    assert "x" in m.genes
    assert rxns.r1.gene_reaction_rule == ""
    assert rxns.r2.gene_reaction_rule == ""
    assert rxns.r3.gene_reaction_rule == "b and c"
    assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
    assert rxns.r5.gene_reaction_rule == "x"
    assert rxns.r6.gene_reaction_rule == "y"
    assert rxns.r7.genes == {m.genes.x, m.genes.z}
    assert rxns.r8.gene_reaction_rule == ""
    remove_genes(m, ["x"], remove_reactions=True)
    assert len(m.reactions) == 7
    assert "r5" not in m.reactions
    assert "x" not in m.genes
    assert rxns.r1.gene_reaction_rule == ""
    assert rxns.r2.gene_reaction_rule == ""
    assert rxns.r3.gene_reaction_rule == "b and c"
    assert rxns.r4.gene_reaction_rule == "(f and b) or (b and c)"
    assert rxns.r6.gene_reaction_rule == "y"
    assert rxns.r7.gene_reaction_rule == "z"
    assert rxns.r7.genes == {m.genes.z}
    assert rxns.r8.gene_reaction_rule == ""



from .annotate import add_SBO
from .delete import (
    delete_model_genes,
    find_gene_knockout_reactions,
    prune_unused_metabolites,
    prune_unused_reactions,
    remove_genes,
    undelete_model_genes,
)
from .modify import (
    escape_ID,
    get_compiled_gene_reaction_rules,
    rename_genes,
)
from .validate import (
    check_mass_balance,
    check_metabolite_compartment_formula,
)



"""Provide functions for phenotype phase plane analysis."""

from itertools import product
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from optlang.interface import OPTIMAL

from ..exceptions import OptimizationError
from ..util import solver as sutil
from .helpers import normalize_cutoff
from .variability import flux_variability_analysis as fva


if TYPE_CHECKING:
    from optlang.interface import Objective

    from cobra import Model, Reaction


def production_envelope(
    model: "Model",
    reactions: List["Reaction"],
    objective: Union[Dict, "Objective", None] = None,
    carbon_sources: Optional[List["Reaction"]] = None,
    points: int = 20,
    threshold: Optional[float] = None,
) -> pd.DataFrame:
    """Calculate the objective value conditioned on all flux combinations.

    The production envelope can be used to analyze a model's ability to
    produce a given compound conditional on the fluxes for another set of
    reactions, such as the uptake rates. The model is alternately optimized
    with respect to minimizing and maximizing the objective and the
    obtained fluxes are recorded. Ranges to compute production is set to the
    effective bounds, i.e., the minimum / maximum fluxes that can be
    obtained given current reaction bounds.

    Parameters
    ----------
    model : cobra.Model
        The model to compute the production envelope for.
    reactions : list of cobra.Reaction
        A list of reaction objects.
    objective : dict or cobra.Model.objective, optional
        The objective (reaction) to use for the production envelope. Use the
        model's current objective if left missing (default None).
    carbon_sources : list of cobra.Reaction, optional
       One or more reactions that are the source of carbon for computing
       carbon (mol carbon in output over mol carbon in input) and mass
       yield (gram product over gram output). Only objectives with a carbon
       containing input and output metabolite is supported. Will identify
       active carbon sources in the medium if none are specified
       (default None).
    points : int, optional
       The number of points to calculate production for (default 20).
    threshold : float, optional
        A cut-off under which flux values will be considered to be zero.
        If not specified, it defaults to `model.tolerance` (default None).

    Returns
    -------
    pandas.DataFrame
        A DataFrame with fixed columns as:
        - carbon_source        : identifiers of carbon exchange reactions
        - flux_maximum         : maximum objective flux
        - flux_minimum         : minimum objective flux
        - carbon_yield_maximum : maximum yield of a carbon source
        - carbon_yield_minimum : minimum yield of a carbon source
        - mass_yield_maximum   : maximum mass yield of a carbon source
        - mass_yield_minimum   : minimum mass yield of a carbon source

        and variable columns (for each input `reactions`) as:
        - reaction_id          : flux at each given point

    Raises
    ------
    ValueError
        If model's objective is comprised of multiple reactions.

    Examples
    --------
    >>> import cobra.test
    >>> from cobra.flux_analysis import production_envelope
    >>> model = cobra.test.create_test_model("textbook")
    >>> production_envelope(model, ["EX_glc__D_e", "EX_o2_e"])
        carbon_source  flux_minimum  carbon_yield_minimum  mass_yield_minimum ...
    0     EX_glc__D_e           0.0                   0.0                 NaN ...
    1     EX_glc__D_e           0.0                   0.0                 NaN ...
    2     EX_glc__D_e           0.0                   0.0                 NaN ...
    3     EX_glc__D_e           0.0                   0.0                 NaN ...
    4     EX_glc__D_e           0.0                   0.0                 NaN ...
    ..            ...           ...                   ...                 ... ...
    395   EX_glc__D_e           NaN                   NaN                 NaN ...
    396   EX_glc__D_e           NaN                   NaN                 NaN ...
    397   EX_glc__D_e           NaN                   NaN                 NaN ...
    398   EX_glc__D_e           NaN                   NaN                 NaN ...
    399   EX_glc__D_e           NaN                   NaN                 NaN ...

    [400 rows x 9 columns]

    """
    reactions = model.reactions.get_by_any(reactions)
    objective = model.solver.objective if objective is None else objective
    data = dict()

    if carbon_sources is None:
        c_input = _find_carbon_sources(model)
    else:
        c_input = model.reactions.get_by_any(carbon_sources)

    if c_input is None:
        data["carbon_source"] = None
    elif hasattr(c_input, "id"):
        data["carbon_source"] = c_input.id
    else:
        data["carbon_source"] = ", ".join(rxn.id for rxn in c_input)

    threshold = normalize_cutoff(model, threshold)

    size = points ** len(reactions)

    for direction in ("minimum", "maximum"):
        data[f"flux_{direction}"] = np.full(size, np.nan, dtype=float)
        data[f"carbon_yield_{direction}"] = np.full(size, np.nan, dtype=float)
        data[f"mass_yield_{direction}"] = np.full(size, np.nan, dtype=float)

    grid = pd.DataFrame(data)

    with model:
        model.objective = objective
        objective_reactions = list(sutil.linear_reaction_coefficients(model))

        if len(objective_reactions) != 1:
            raise ValueError(
                "Cannot calculate yields for objectives with multiple reactions."
            )
        c_output = objective_reactions[0]
        min_max = fva(model, reactions, fraction_of_optimum=0)
        min_max[min_max.abs() < threshold] = 0.0
        points = list(
            product(
                *[
                    np.linspace(
                        min_max.at[rxn.id, "minimum"],
                        min_max.at[rxn.id, "maximum"],
                        points,
                        endpoint=True,
                    )
                    for rxn in reactions
                ]
            )
        )
        tmp = pd.DataFrame(points, columns=[rxn.id for rxn in reactions])
        grid = pd.concat([grid, tmp], axis=1, copy=False)
        _add_envelope(model, reactions, grid, c_input, c_output, threshold)

    return grid


def _add_envelope(
    model: "Model",
    reactions: List["Reaction"],
    grid: pd.DataFrame,
    c_input: List["Reaction"],
    c_output: List["Reaction"],
    threshold: float,
) -> None:
    """Add a production envelope based on the parameters provided.

    Parameters
    ----------
    model : cobra.Model
        The model to operate on.
    reactions : list of cobra.Reaction
        The input reaction objects.
    grid : pandas.DataFrame
        The DataFrame containing all the data regarding the operation.
    c_input : list of cobra.Reaction
        The list of reaction objects acting as carbon inputs.
    c_output : list of cobra.Reaction
        The list of reaction objects acting as carbon outputs.

    """
    if c_input is not None:
        input_components = [_reaction_elements(rxn) for rxn in c_input]
        output_components = _reaction_elements(c_output)

        try:
            input_weights = [_reaction_weight(rxn) for rxn in c_input]
            output_weight = _reaction_weight(c_output)
        except ValueError:
            input_weights = []
            output_weight = []
    else:
        input_components = []
        output_components = []
        input_weights = []
        output_weight = []

    for direction in ("minimum", "maximum"):
        with model:
            model.objective_direction = direction

            for i in range(len(grid)):
                with model:
                    for rxn in reactions:
                        point = grid.at[i, rxn.id]
                        rxn.bounds = point, point
                    obj_val = model.slim_optimize()

                    if model.solver.status != OPTIMAL:
                        continue

                    grid.at[i, f"flux_{direction}"] = (
                        0.0 if np.abs(obj_val) < threshold else obj_val
                    )

                    if c_input is not None:
                        grid.at[i, f"carbon_yield_{direction}"] = _total_yield(
                            [rxn.flux for rxn in c_input],
                            input_components,
                            obj_val,
                            output_components,
                        )
                        grid.at[i, f"mass_yield_{direction}"] = _total_yield(
                            [rxn.flux for rxn in c_input],
                            input_weights,
                            obj_val,
                            output_weight,
                        )


def _total_yield(
    input_fluxes: List[float],
    input_elements: List[float],
    output_flux: List[float],
    output_elements: List[float],
) -> float:
    """Compute total output per input unit.

    Units are typically mol carbon atoms or gram of source and product.

    Parameters
    ----------
    input_fluxes : list of float
        A list of input reaction fluxes in the same order as the
        `input_components`.
    input_elements : list of float
        A list of reaction components which are in turn list of numbers.
    output_flux : float
        The output flux value.
    output_elements : list
        A list of stoichiometrically weighted output reaction components.

    Returns
    -------
    float
        The ratio between output (mol carbon atoms or grams of product) and
        input (mol carbon atoms or grams of source compounds). If input flux
        of carbon sources is zero then numpy.nan is returned.

    """
    carbon_input_flux = sum(
        _total_components_flux(flux, components, consumption=True)
        for flux, components in zip(input_fluxes, input_elements)
    )
    carbon_output_flux = _total_components_flux(
        output_flux, output_elements, consumption=False
    )
    try:
        return carbon_output_flux / carbon_input_flux
    except ZeroDivisionError:
        return np.nan


def _reaction_elements(reaction: "Reaction") -> List[float]:
    """Split metabolites into atoms times their stoichiometric coefficients.

    Parameters
    ----------
    reaction : cobra.Reaction
        The reaction whose metabolite components are desired.

    Returns
    -------
    list of float
        Each of the reaction's metabolites' desired carbon elements (if any)
        times that metabolite's stoichiometric coefficient.

    """
    c_elements = [
        coeff * met.elements.get("C", 0) for met, coeff in reaction.metabolites.items()
    ]
    return [elem for elem in c_elements if elem != 0]


def _reaction_weight(reaction: "Reaction") -> List[float]:
    """Return the metabolite weight times its stoichiometric coefficient.

    Parameters
    ----------
    reaction : cobra.Reaction
        The reaction whose metabolite component weights is desired.

    Returns
    -------
    list of float
        Each of reaction's metabolite components' weights.

    Raises
    ------
    ValueError
        If more than one metabolite comprises the `reaction`.

    """
    if len(reaction.metabolites) != 1:
        raise ValueError(
            "Reaction weight is only defined for single "
            "metabolite products or educts."
        )

    met, coeff = next(iter(reaction.metabolites.items()))
    return [coeff * met.formula_weight]


def _total_components_flux(
    flux: float, components: List[float], consumption: bool = True
) -> float:
    """Compute the total components consumption or production flux.

    Parameters
    ----------
    flux : float
        The reaction flux for the components.
    components : list of float
        List of stoichiometrically weighted components.
    consumption : bool, optional
        Whether to sum up consumption or production fluxes (default True).

    Returns
    -------
    float
        The total consumption or production flux of `components`.

    """
    direction = 1 if consumption else -1
    c_flux = [elem * flux * direction for elem in components]
    return sum([flux for flux in c_flux if flux > 0])


def _find_carbon_sources(model: "Model") -> List["Reaction"]:
    """Find all active carbon source reactions.

    Parameters
    ----------
    model : Model
        The model whose active carbon sources need to found.

    Returns
    -------
    list of cobra.Reaction
       The medium reactions with carbon input flux.

    """
    try:
        model.slim_optimize(error_value=None)
    except OptimizationError:
        return []

    reactions = model.reactions.get_by_any(list(model.medium))
    reactions_fluxes = [
        (
            rxn,
            _total_components_flux(rxn.flux, _reaction_elements(rxn), consumption=True),
        )
        for rxn in reactions
    ]
    return [rxn for rxn, c_flux in reactions_fluxes if c_flux > 0]



"""Provide the model summary class."""


import logging
from operator import attrgetter
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import pandas as pd

from cobra.core import Reaction
from cobra.flux_analysis import flux_variability_analysis, pfba
from cobra.summary import Summary
from cobra.util.solver import linear_reaction_coefficients


if TYPE_CHECKING:
    from cobra.core import Metabolite, Model, Solution


logger = logging.getLogger(__name__)


class ModelSummary(Summary):
    """
    Define the model summary.

    Attributes
    ----------
    uptake_flux : pandas.DataFrame
        A pandas DataFrame of only the uptake fluxes.
    secretion_flux : pandas.DataFrame
        A pandas DataFrame of only the consuming fluxes.
    See Also
    --------
    Summary : Parent that defines further attributes.
    MetaboliteSummary
    ReactionSummary

    """

    def __init__(
        self,
        *,
        model: "Model",
        solution: Optional["Solution"] = None,
        fva: Optional[Union[float, pd.DataFrame]] = None,
        **kwargs,
    ):
        """
        Initialize a model summary.

        Parameters
        ----------
        model : cobra.Model
            The metabolic model for which to generate a metabolite summary.
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution (default None).
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched (default None).

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        See Also
        --------
        Summary : Parent that has further default parameters.
        MetaboliteSummary
        ReactionSummary

        """
        super().__init__(**kwargs)
        self._objective = None
        self._objective_value = None
        self._boundary = None
        self._boundary_metabolites = None
        self.uptake_flux: Optional[pd.DataFrame] = None
        self.secretion_flux: Optional[pd.DataFrame] = None
        self._generate(model, solution, fva)

    def _generate(
        self,
        model: "Model",
        solution: Optional["Solution"],
        fva: Optional[Union[float, pd.DataFrame]],
    ) -> None:
        """
        Prepare the data for the summary instance.

        Parameters
        ----------
        model : cobra.Model
            The metabolic model for which to generate a metabolite summary.
        solution : cobra.Solution, optional
            A previous model solution to use for generating the summary. If
            ``None``, the summary method will generate a parsimonious flux
            distribution.
        fva : pandas.DataFrame or float, optional
            Whether or not to include flux variability analysis in the output.
            If given, `fva` should either be a previous FVA solution matching the
            model or a float between 0 and 1 representing the fraction of the
            optimum objective to be searched.

        """
        super()._generate(model=model, solution=solution, fva=fva)

        coefficients = linear_reaction_coefficients(model)
        if solution is None:
            logger.info("Generating new parsimonious flux distribution.")
            solution = pfba(model)

        if isinstance(fva, float):
            logger.info("Performing flux variability analysis.")
            fva = flux_variability_analysis(
                model=model,
                reaction_list=model.boundary,
                fraction_of_optimum=fva,
            )
        if coefficients:
            self._objective: Dict["Reaction", float] = {
                rxn.copy(): coef for rxn, coef in coefficients.items()
            }
            self._objective_value: float = sum(
                solution[rxn.id] * coef for rxn, coef in self._objective.items()
            )
        else:
            logger.warning(
                "Non-linear or non-reaction model objective. Falling back to minimal "
                "display."
            )
            self._objective = {
                Reaction(id="Expression", name="Expression"): float("nan")
            }
            self._objective_value: float = float("nan")
        self._boundary: List["Reaction"] = [
            rxn.copy() for rxn in sorted(model.boundary, key=attrgetter("id"))
        ]
        self._boundary_metabolites: List["Metabolite"] = [
            met.copy() for rxn in self._boundary for met in rxn.metabolites
        ]
        flux = pd.DataFrame(
            data=[
                (rxn.id, met.id, rxn.get_coefficient(met.id), solution[rxn.id])
                for rxn, met in zip(self._boundary, self._boundary_metabolites)
            ],
            columns=["reaction", "metabolite", "factor", "flux"],
            index=[r.id for r in self._boundary],
        )
        # Scale fluxes by stoichiometric coefficient.
        flux["flux"] *= flux["factor"]

        if fva is not None:
            flux = flux.join(fva)
            view = flux[["flux", "minimum", "maximum"]]
            # Set fluxes below model tolerance to zero.
            flux[["flux", "minimum", "maximum"]] = view.where(
                view.abs() >= model.tolerance, 0
            )
            # Create the scaled compound flux.
            flux[["minimum", "maximum"]] = flux[["minimum", "maximum"]].mul(
                flux["factor"], axis=0
            )
            # Negative factors invert the minimum/maximum relationship.
            negative = flux["factor"] < 0
            tmp = flux.loc[negative, "maximum"]
            flux.loc[negative, "maximum"] = flux.loc[negative, "minimum"]
            flux.loc[negative, "minimum"] = tmp
            # Add zero to turn negative zero into positive zero for nicer display later.
            flux[["flux", "minimum", "maximum"]] += 0
        else:
            # Set fluxes below model tolerance to zero.
            flux.loc[flux["flux"].abs() < model.tolerance, "flux"] = 0
            # Add zero to turn negative zero into positive zero for nicer display later.
            flux["flux"] += 0

        # Create production table from producing fluxes or zero fluxes where the
        # metabolite is a product in the reaction.
        is_produced = (flux["flux"] > 0) | ((flux["flux"] == 0) & (flux["factor"] > 0))
        if fva is not None:
            self.uptake_flux = flux.loc[
                is_produced, ["flux", "minimum", "maximum", "reaction", "metabolite"]
            ].copy()
        else:
            self.uptake_flux = flux.loc[
                is_produced, ["flux", "reaction", "metabolite"]
            ].copy()

        # Create consumption table from consuming fluxes or zero fluxes where the
        # metabolite is a substrate in the reaction.
        is_consumed = (flux["flux"] < 0) | ((flux["flux"] == 0) & (flux["factor"] < 0))
        if fva is not None:
            self.secretion_flux = flux.loc[
                is_consumed, ["flux", "minimum", "maximum", "reaction", "metabolite"]
            ].copy()
        else:
            self.secretion_flux = flux.loc[
                is_consumed, ["flux", "reaction", "metabolite"]
            ].copy()

        self._flux = flux

    def _display_flux(
        self, frame: pd.DataFrame, names: bool, element: str, threshold: float
    ) -> pd.DataFrame:
        """
        Transform a flux data frame for display.

        Parameters
        ----------
        frame : pandas.DataFrame
            Either the producing or the consuming fluxes.
        names : bool
            Whether or not elements should be displayed by their common names.
        element : str
            The atomic element to summarize fluxes for.
        threshold : float
            Hide fluxes below the threshold from being displayed.

        Returns
        -------
        pandas.DataFrame
            The transformed pandas DataFrame with flux percentages and reaction
            definitions.

        """
        if "minimum" in frame.columns and "maximum" in frame.columns:
            frame = frame.loc[
                (frame["flux"].abs() >= threshold)
                | (frame["minimum"].abs() >= threshold)
                | (frame["maximum"].abs() >= threshold),
                :,
            ].copy()
        else:
            frame = frame.loc[frame["flux"].abs() >= threshold, :].copy()

        metabolites = {m.id: m for m in self._boundary_metabolites}

        element_num = f"{element}-Number"
        frame[element_num] = [
            metabolites[met_id].elements.get(element, 0)
            for met_id in frame["metabolite"]
        ]
        element_percent = f"{element}-Flux"
        frame[element_percent] = frame[element_num] * frame["flux"].abs()
        total = frame[element_percent].sum()
        if total > 0.0:
            frame[element_percent] /= total
        frame[element_percent] = [f"{x:.2%}" for x in frame[element_percent]]

        if names:
            frame["metabolite"] = [
                metabolites[met_id].name for met_id in frame["metabolite"]
            ]

        if "minimum" in frame.columns and "maximum" in frame.columns:
            frame["range"] = list(
                frame[["minimum", "maximum"]].itertuples(index=False, name=None)
            )
            return frame[
                [
                    "metabolite",
                    "reaction",
                    "flux",
                    "range",
                    element_num,
                    element_percent,
                ]
            ]
        else:
            return frame[
                ["metabolite", "reaction", "flux", element_num, element_percent]
            ]

    @staticmethod
    def _string_table(frame: pd.DataFrame, float_format: str, column_width: int) -> str:
        """
        Create a pretty string representation of the data frame.

        Parameters
        ----------
        frame : pandas.DataFrame
            A pandas DataFrame of fluxes.
        float_format : str
            Format string for floats.
        column_width : int
            The maximum column width for each row.

        Returns
        -------
        str
            The data frame formatted as a pretty string.

        """
        frame.columns = [header.title() for header in frame.columns]
        return frame.to_string(
            header=True,
            index=False,
            na_rep="",
            formatters={
                "Flux": f"{{:{float_format}}}".format,
                "Range": lambda pair: f"[{pair[0]:{float_format}}; "
                f"{pair[1]:{float_format}}]",
            },
            max_colwidth=column_width,
        )

    @staticmethod
    def _html_table(frame: pd.DataFrame, float_format: str) -> str:
        """
        Create an HTML representation of the data frame.

        Parameters
        ----------
        frame : pandas.DataFrame
            A pandas DataFrame of fluxes.
        float_format : str
            Format string for floats.

        Returns
        -------
        str
            The data frame formatted as HTML.

        """
        frame.columns = [header.title() for header in frame.columns]
        return frame.to_html(
            header=True,
            index=False,
            na_rep="",
            formatters={
                "Flux": f"{{:{float_format}}}".format,
                "Range": lambda pair: f"[{pair[0]:{float_format}}; "
                f" {pair[1]:{float_format}}]",
            },
        )

    def _string_objective(self, names: bool) -> str:
        """
        Return a string representation of the objective.

        Parameters
        ----------
        names : bool, optional
            Whether or not elements should be displayed by their common names.

        Returns
        -------
        str
            The objective expression and value as a string.

        """
        if names:
            objective = " + ".join(
                [f"{coef} {rxn.name}" for rxn, coef in self._objective.items()]
            )
        else:
            objective = " + ".join(
                [f"{coef} {rxn.id}" for rxn, coef in self._objective.items()]
            )
        return f"{objective} = {self._objective_value}"

    def to_string(
        self,
        names: bool = False,
        element: str = "C",
        threshold: Optional[float] = None,
        float_format: str = ".4G",
        column_width: int = 79,
    ) -> str:
        """
        Return a pretty string representation of the model summary.

        Parameters
        ----------
        names : bool, optional
            Whether or not elements should be displayed by their common names
            (default False).
        element : str, optional
            The atomic element to summarize uptake and secretion for (default 'C').
        threshold : float, optional
            Hide fluxes below the threshold from being displayed. If no value is
            given, the model tolerance is used (default None).
        float_format : str, optional
            Format string for floats (default '.4G').
        column_width : int, optional
            The maximum column width for each row (default 79).

        Returns
        -------
        str
            The summary formatted as a pretty string.

        """
        threshold = self._normalize_threshold(threshold)

        objective = self._string_objective(names)

        uptake = self._string_table(
            self._display_flux(self.uptake_flux, names, element, threshold),
            float_format,
            column_width,
        )

        secretion = self._string_table(
            self._display_flux(self.secretion_flux, names, element, threshold),
            float_format,
            column_width,
        )

        return (
            f"Objective\n"
            f"=========\n"
            f"{objective}\n\n"
            f"Uptake\n"
            f"------\n"
            f"{uptake}\n\n"
            f"Secretion\n"
            f"---------\n"
            f"{secretion}\n"
        )

    def to_html(
        self,
        names: bool = False,
        element: str = "C",
        threshold: Optional[float] = None,
        float_format: str = ".4G",
    ) -> str:
        """
        Return a rich HTML representation of the model summary.

        Parameters
        ----------
        names : bool, optional
            Whether or not elements should be displayed by their common names
            (default False).
        element : str, optional
            The atomic element to summarize uptake and secretion for (default 'C').
        threshold : float, optional
            Hide fluxes below the threshold from being displayed. If no value is
            given, the model tolerance is used (default None).
        float_format : str, optional
            Format string for floats (default '.4G').

        Returns
        -------
        str
            The summary formatted as HTML.

        """
        threshold = self._normalize_threshold(threshold)

        objective = self._string_objective(names)

        uptake = self._html_table(
            self._display_flux(self.uptake_flux, names, element, threshold),
            float_format,
        )

        secretion = self._html_table(
            self._display_flux(self.secretion_flux, names, element, threshold),
            float_format,
        )

        return (
            f"<h3>Objective</h3>"
            f"<p>{objective}</p>"
            f"<h4>Uptake</h4>"
            f"{uptake}"
            f"<h4>Secretion</h4>"
            f"{secretion}"
        )



"""Test functionalities of model component modifcations."""

from cobra.core import Model
from cobra.manipulation import escape_ID, rename_genes


def test_escape_ids(model: Model) -> None:
    """Test model component IDs' SBML compliance."""
    model.reactions.PGI.gene_reaction_rule = "a.b or c"
    assert "a.b" in model.genes
    escape_ID(model)
    assert "a.b" not in model.genes


def test_rename_genes(model: Model) -> None:
    """Test gene renaming functionality."""
    original_name = model.genes.b1241.name
    rename_dict = {
        "b1241": "foo",
        "hello": "world",
        "b3115": "b3115",
        "b2465": "b3919",
        "bar": "2935",
    }
    rename_genes(model, rename_dict)
    for i in rename_dict.keys():
        if i not in rename_dict.values():
            assert i not in model.genes

    assert "b3115" in model.genes
    assert "foo" in model.genes
    assert "world" not in model.genes
    # make sure the object name was preserved
    assert model.genes.foo.name == original_name
    # make sure the reactions are correct
    assert len(model.genes.foo.reactions) == 2
    assert model.reactions.ACALD.gene_reaction_rule == "b0351 or foo"
    assert model.reactions.TPI.gene_reaction_rule == "b3919"
    assert model.reactions.TPI.genes == {model.genes.b3919}
    assert model.reactions.TKT1.gene_reaction_rule == "b2935 or b3919"
    assert model.reactions.TKT1.genes == {model.genes.b2935, model.genes.b3919}
    assert model.genes.b3919.reactions == {
        model.reactions.get_by_id(i) for i in ("TKT1", "TKT2", "TPI")
    }



"""Provide a function ``load_model`` to access remote model repositories."""


import gzip
import logging
from typing import TYPE_CHECKING, Iterable

import diskcache
import httpx
import libsbml

from ...core import Configuration
from ..sbml import _sbml_to_model
from .abstract_model_repository import AbstractModelRepository
from .bigg_models_repository import BiGGModels
from .biomodels_repository import BioModels


if TYPE_CHECKING:
    from cobra.core import Model


logger = logging.getLogger(__name__)
configuration = Configuration()


def load_model(
    model_id: str,
    repositories: Iterable[AbstractModelRepository] = (BiGGModels(), BioModels()),
    cache: bool = True,
) -> "Model":
    """
    Download an SBML model from a remote repository.

    Downloaded SBML documents are by default stored in a cache on disk such that future
    access is much faster. By default, models can be loaded from the following
    repositories:

    * BiGG Models
    * BioModels

    You can use the ``AbstractModelRepository`` class as a parent to implement your own
    repository accessor which you pass to the ``load_model`` function. In case you
    implement a new interface, please consider submitting a pull request to COBRApy.

    Parameters
    ----------
    model_id : str
        The identifier of the desired metabolic model. This is typically repository
        specific.
    repositories : iterable, optional
        An iterable of repository accessor instances. The model_id is searched in order.
    cache : bool, optional
        Whether or not to use the local caching mechanism (default yes).

    Returns
    -------
    Model
        A model instance generated from the SBML document.

    Raises
    ------
    RuntimeError
        As with any internet connection, there are multiple errors that can occur.

    Examples
    --------
    # Most of the time calling `load_model` with an identifier should be enough.
    >>> print(load_model("e_coli_core"))
    e_coli_core
    >>> print(load_model("MODEL1510010000"))
    MODEL1510010000

    See Also
    --------
    BiGGModels
    BioModels

    """
    if cache:
        data = _cached_load(
            model_id=model_id,
            repositories=repositories,
        )
    else:
        data = _fetch_model(model_id=model_id, repositories=repositories)
    return get_model_from_gzip_sbml(data)


def _cached_load(
    model_id: str,
    repositories: Iterable[AbstractModelRepository],
) -> bytes:
    """
    Attempt to load a gzip-compressed SBML document from the cache.

    If the given model identifier is not in the cache, the remote repositories are
    searched.

    Parameters
    ----------
    model_id : str
        The identifier of the desired metabolic model. This is typically repository
        specific.
    repositories : iterable
        An iterable of repository accessor instances. The model_id is searched in order.

    Returns
    -------
    bytes
        A gzip-compressed, UTF-8 encoded SBML document.

    """
    with diskcache.Cache(
        directory=str(configuration.cache_directory),
        size_limit=configuration.max_cache_size,
    ) as cache:
        try:
            return cache[model_id]
        except KeyError:
            data = _fetch_model(model_id=model_id, repositories=repositories)
            cache.set(key=model_id, value=data, expire=configuration.cache_expiration)
            return data


def _fetch_model(
    model_id: str,
    repositories: Iterable[AbstractModelRepository],
) -> bytes:
    """
    Attempt to load a gzip-compressed SBML document from the given repositories.

    Parameters
    ----------
    model_id : str
        The identifier of the desired metabolic model. This is typically repository
        specific.
    repositories : iterable
        An iterable of repository accessor instances. The model_id is searched in order.

    Returns
    -------
    bytes
        A gzip-compressed, UTF-8 encoded SBML document.

    """
    for repository in repositories:
        logger.info(
            f"Attempting to fetch '{model_id}' from the {repository.name} repository."
        )
        try:
            return repository.get_sbml(model_id=model_id)
        except httpx.HTTPStatusError as error:
            if error.response.status_code == 404:
                logger.debug(
                    f"Model '{model_id}' not found in the {repository.name} repository."
                )
                continue
            raise RuntimeError(
                f"The connection to the {repository.name} repository failed."
            ) from error
        except httpx.RequestError as error:
            raise RuntimeError(
                f"The connection to the {repository.name} repository failed."
            ) from error
    raise RuntimeError(
        f"The model '{model_id}' could not be found in any of the repositories."
    )


def get_model_from_gzip_sbml(stream: bytes) -> "Model":
    """
    Generate a model instance from a gzip-compressed, UTF-8 encoded SBML document.

    Parameters
    ----------
    stream : bytes
        A gzip-compressed, UTF-8 encoded SBML document.

    Returns
    -------
    Model
        A model instance generated from the SBML document.

    """
    return _sbml_to_model(
        libsbml.readSBMLFromString(gzip.decompress(stream).decode("utf-8"))
    )



"""Module for shared exceptions in the Cobra package."""
import optlang.interface


class OptimizationError(Exception):
    """Exception for Optimization issues."""

    def __init__(self, message):
        """Inherit parent behaviors."""
        super(OptimizationError, self).__init__(message)


class Infeasible(OptimizationError):
    """Exception for Infeasible issues."""

    pass


class Unbounded(OptimizationError):
    """Exception for Unbounded issues."""

    pass


class FeasibleButNotOptimal(OptimizationError):
    """Exception for Non-Optimal issues."""

    pass


class UndefinedSolution(OptimizationError):
    """Exception for Undefined issues."""

    pass


class SolverNotFound(Exception):
    """A simple Exception when a solver can not be found."""

    pass


OPTLANG_TO_EXCEPTIONS_DICT = dict(
    (
        (optlang.interface.INFEASIBLE, Infeasible),
        (optlang.interface.UNBOUNDED, Unbounded),
        (optlang.interface.FEASIBLE, FeasibleButNotOptimal),
        (optlang.interface.UNDEFINED, UndefinedSolution),
    )
)



"""Provide functions and helpers to obtain minimal growth media."""

import logging
from typing import TYPE_CHECKING, Iterable, Union

import pandas as pd
from optlang.interface import OPTIMAL
from optlang.symbolics import Zero

from .boundary_types import find_boundary_types


if TYPE_CHECKING:
    from cobra import Model, Reaction


logger = logging.getLogger(__name__)


def add_linear_obj(model: "Model") -> None:
    r"""Add a linear version of a minimal medium to the model solver.

    Changes the optimization objective to finding the growth medium
    requiring the smallest total import flux:

    ..math:: minimize \sum_{r_i in import_reactions} |r_i|

    Parameters
    ----------
    model : cobra.Model
        The cobra model to modify.

    """
    coefs = {}
    for rxn in find_boundary_types(model, "exchange"):
        export = len(rxn.reactants) == 1
        if export:
            coefs[rxn.reverse_variable] = 1
        else:
            coefs[rxn.forward_variable] = 1
    model.objective.set_linear_coefficients(coefs)
    model.objective.direction = "min"


def add_mip_obj(model: "Model") -> None:
    """Add a mixed-integer version of a minimal medium to the model.

    Changes the optimization objective to finding the medium with the least
    components:

    minimize size(R) where R part of import_reactions

    Arguments
    ---------
    model : cobra.model
        The model to modify.

    """
    if len(model.variables) > 1e4:
        logger.warning(
            "The MIP version of minimal media is extremely slow for "
            "models that large :("
        )
    exchange_rxns = find_boundary_types(model, "exchange")
    big_m = max(abs(b) for r in exchange_rxns for b in r.bounds)
    prob = model.problem
    coefs = {}
    to_add = []
    for rxn in exchange_rxns:
        export = len(rxn.reactants) == 1
        indicator = prob.Variable("ind_" + rxn.id, lb=0, ub=1, type="binary")
        if export:
            vrv = rxn.reverse_variable
            indicator_const = prob.Constraint(
                vrv - indicator * big_m, ub=0, name="ind_constraint_" + rxn.id
            )
        else:
            vfw = rxn.forward_variable
            indicator_const = prob.Constraint(
                vfw - indicator * big_m, ub=0, name="ind_constraint_" + rxn.id
            )
        to_add.extend([indicator, indicator_const])
        coefs[indicator] = 1
    model.add_cons_vars(to_add)
    model.solver.update()
    model.objective.set_linear_coefficients(coefs)
    model.objective.direction = "min"


def _as_medium(
    exchanges: Iterable["Reaction"], tolerance: float = 1e-6, exports: bool = False
) -> pd.Series:
    """Convert a solution to medium.

    Parameters
    ----------
    exchanges : list of cobra.reaction
        The exchange reactions to consider.
    tolerance : float > 0, optional
        The absolute tolerance for fluxes. Fluxes with an absolute value
        smaller than this number will be ignored (default 1e-6).
    exports : bool, optional
        Whether to return export fluxes as well (default False).

    Returns
    -------
    pandas.Series
        The "medium", meaning all active import fluxes in the solution.

    """
    logger.debug("Formatting medium.")
    medium = pd.Series()
    for rxn in exchanges:
        export = len(rxn.reactants) == 1
        flux = rxn.flux
        if abs(flux) < tolerance:
            continue
        if export:
            medium[rxn.id] = -flux
        elif not export:
            medium[rxn.id] = flux
    if not exports:
        medium = medium[medium > 0]

    return medium


def minimal_medium(
    model: "Model",
    min_objective_value: float = 0.1,
    exports: bool = False,
    minimize_components: Union[bool, int] = False,
    open_exchanges: bool = False,
) -> Union[pd.Series, pd.DataFrame, None]:
    """Find the minimal growth medium for the `model`.

    Finds the minimal growth medium for the `model` which allows for
    model as well as individual growth. Here, a minimal medium can either
    be the medium requiring the smallest total import flux or the medium
    requiring the least components (ergo ingredients), which will be much
    slower due to being a mixed integer problem (MIP).

    Parameters
    ----------
    model : cobra.model
        The model to modify.
    min_objective_value : float > 0 or array-like object, optional
        The minimum growth rate (objective) that has to be achieved
        (default 0.1).
    exports : bool, optional
        Whether to include export fluxes in the returned medium.
        Defaults to False which will only return import fluxes
        (default False).
    minimize_components : bool or int > 0, optional
        Whether to minimize the number of components instead of the total
        import flux. Might be more intuitive if set to True, but may also
        be slow to calculate for large communities. If set to a number `n`
        will return up to `n` alternative solutions all with the same
        number of components (default False).
    open_exchanges : bool or number, optional
        Whether to ignore currently set bounds and make all exchange
        reactions in the `model` possible. If set to a `number`, all
        exchange reactions will be opened with (-`number`, `number`) as
        bounds (default False).

    Returns
    -------
    pandas.Series, pandas.DataFrame or None
        A pandas.Series giving the import flux for each required import
        reaction and (optionally) the associated export fluxes. All exchange
        fluxes are oriented into the import reaction e.g. positive fluxes
        denote imports and negative fluxes exports. If `minimize_components`
        is a number larger than 1, may return a pandas.DataFrame where each
        column is a minimal medium. Returns None, if the minimization is
        infeasible (for instance if min_growth > maximum growth rate).

    Notes
    -----
    Due to numerical issues, the `minimize_components` option will usually
    only minimize the number of "large" import fluxes. Specifically, the
    detection limit is given by ``integrality_tolerance * max_bound`` where
    ``max_bound`` is the largest bound on an import reaction. Thus, if you
    are interested in small import fluxes as well you may have to adjust the
    solver tolerance at first with `model.tolerance = 1e-7` for instance.
    However, this will be *very* slow for large models especially with GLPK.

    """
    exchange_rxns = find_boundary_types(model, "exchange")
    if isinstance(open_exchanges, bool):
        open_bound = 1000
    else:
        open_bound = open_exchanges

    with model as mod:
        if open_exchanges:
            logger.debug(f"Opening exchanges for {len(exchange_rxns)} imports.")
            for rxn in exchange_rxns:
                rxn.bounds = (-open_bound, open_bound)
        logger.debug("Applying objective value constraints.")
        obj_const = mod.problem.Constraint(
            mod.objective.expression,
            lb=min_objective_value,
            name="medium_obj_constraint",
        )
        mod.add_cons_vars([obj_const])
        mod.solver.update()
        mod.objective = Zero
        logger.debug("Adding new media objective.")
        tol = mod.solver.configuration.tolerances.feasibility

        if minimize_components:
            add_mip_obj(mod)
            if isinstance(minimize_components, bool):
                minimize_components = 1
            seen = set()
            best = num_components = mod.slim_optimize()
            if mod.solver.status != OPTIMAL:
                logger.warning("Minimization of medium was infeasible.")
                return None
            exclusion = mod.problem.Constraint(Zero, ub=0)
            mod.add_cons_vars([exclusion])
            mod.solver.update()
            media = []

            for i in range(minimize_components):
                logger.info(f"Finding alternative medium #{(i + 1)}.")
                vars = [mod.variables["ind_" + s] for s in seen]
                if len(seen) > 0:
                    exclusion.set_linear_coefficients(dict.fromkeys(vars, 1))
                    exclusion.ub = best - 1
                num_components = mod.slim_optimize()
                if mod.solver.status != OPTIMAL or num_components > best:
                    if i == 0:
                        logger.warning(
                            "Could not get an optimal solution. "
                            "This is usually due to numerical instability. "
                            "Possible remedies are relaoding the model "
                            "from scratch, switching to a different solver, "
                            "or decreasing the solver tolerance. Please, "
                            "carefully read the note on numerical instability "
                            "in the function documentation."
                        )
                        return None
                    break
                medium = _as_medium(exchange_rxns, tol, exports=exports)
                media.append(medium)
                seen.update(medium[medium > 0].index)
            if len(media) > 1:
                medium = pd.concat(media, axis=1, sort=True).fillna(0.0)
                medium.sort_index(axis=1, inplace=True)
            else:
                medium = media[0]
        else:
            add_linear_obj(mod)
            mod.slim_optimize()
            if mod.solver.status != OPTIMAL:
                logger.warning("Minimization of medium was infeasible.")
                return None
            medium = _as_medium(exchange_rxns, tol, exports=exports)

    return medium



"""Provide functions for loading metabolic models over the wire."""


import gzip
from io import BytesIO
from typing import List

import httpx
import pydantic

from .abstract_model_repository import AbstractModelRepository


class BioModelsFile(pydantic.BaseModel):
    """Define a single BioModels file description."""

    name: str
    size: int = pydantic.Field(alias="fileSize")


class BioModelsFilesResponse(pydantic.BaseModel):
    """Define the BioModels files JSON response."""

    main: List[BioModelsFile] = []


class BioModels(AbstractModelRepository):
    """
    Define a concrete implementation of the BioModels repository.

    Attributes
    ----------
    name : str
        The name of the BioModels repository.

    """

    name: str = "BioModels"

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """
        Initialize a BioModels repository interface.

        Other Parameters
        ----------------
        kwargs
            Passed to the parent constructor in order to enable multiple inheritance.

        """
        super().__init__(url="https://www.ebi.ac.uk/biomodels/model/", **kwargs)

    def get_sbml(self, model_id: str) -> bytes:
        """
        Attempt to download an SBML document from the repository.

        Parameters
        ----------
        model_id : str
            The identifier of the desired metabolic model. This is typically repository
            specific.

        Returns
        -------
        bytes
            A gzip-compressed, UTF-8 encoded SBML document.

        Raises
        ------
        httpx.HTTPError
            In case there are any connection problems.

        """
        data = BytesIO()
        response = httpx.get(
            url=self._url.join(f"files/{model_id}"),
            headers={"Accept": "application/json"},
        )
        response.raise_for_status()
        files = BioModelsFilesResponse.parse_obj(response.json())
        for model in files.main:
            if model.name.endswith("xml"):
                break
        else:
            RuntimeError(f"Could not find an SBML document for '{model_id}'.")
        with self._progress, httpx.stream(
            method="GET",
            url=self._url.join(f"download/{model_id}"),
            params={"filename": model.name},
        ) as response:
            response.raise_for_status()
            task_id = self._progress.add_task(
                description="download",
                total=model.size,
                model_id=model_id,
            )
            for chunk in response.iter_bytes():
                data.write(chunk)
                self._progress.update(task_id=task_id, advance=len(chunk))
        data.seek(0)
        return gzip.compress(data.read())



from .deletion import (
    double_gene_deletion,
    double_reaction_deletion,
    single_gene_deletion,
    single_reaction_deletion,
)
from .fastcc import fastcc
from .gapfilling import gapfill
from .geometric import geometric_fba
from .loopless import loopless_solution, add_loopless
from .moma import add_moma, moma
from .parsimonious import pfba
from .variability import (
    find_blocked_reactions,
    find_essential_genes,
    find_essential_reactions,
    flux_variability_analysis,
)
from .phenotype_phase_plane import production_envelope
from .room import add_room, room



"""Provide the base class and associated functions for Hit-and-Run samplers."""

import ctypes
import logging
from abc import ABC, abstractmethod
from multiprocessing import Array
from time import time
from typing import TYPE_CHECKING, NamedTuple, Optional, Tuple

import numpy as np
import pandas as pd
from optlang.interface import OPTIMAL
from optlang.symbolics import Zero

from cobra.util import constraint_matrices, create_stoichiometric_matrix, nullspace


if TYPE_CHECKING:
    from cobra import Model


logger = logging.getLogger(__name__)


Problem = NamedTuple(
    "Problem",
    [
        ("equalities", np.ndarray),
        ("b", np.ndarray),
        ("inequalities", np.ndarray),
        ("bounds", np.ndarray),
        ("variable_fixed", np.ndarray),
        ("variable_bounds", np.ndarray),
        ("nullspace", np.matrix),
        ("homogeneous", bool),
    ],
)
"""Define the matrix representation of a sampling problem.

A named tuple consisting of 6 arrays, 1 matrix and 1 boolean.

Attributes
----------
equalities : numpy.array
    All equality constraints in the model.
b : numpy.array
    The right side of the equality constraints.
inequalities : numpy.array
    All inequality constraints in the model.
bounds : numpy.array
    The lower and upper bounds for the inequality constraints.
variable_fixed : numpy.array
    A boolean vector indicating whether the variable at that index is
    fixed i.e., whether `variable.lower_bound == variable.upper_bound`.
variable_bounds : numpy.array
    The lower and upper bounds for the variables.
nullspace : numpy.matrix
    A matrix containing the nullspace of the equality constraints.
    Each column is one basis vector.
homogeneous: bool
    Indicates whether the sampling problem is homogeneous, e.g. whether
    there exist no non-zero fixed variables or constraints.

"""


def shared_np_array(
    shape: Tuple[int, int], data: Optional[np.ndarray] = None, integer: bool = False
) -> np.ndarray:
    """Create a new numpy array that resides in shared memory.

    Parameters
    ----------
    shape : tuple of int
        The shape of the new array.
    data : numpy.array, optional
        Data to copy to the new array. Has to have the same shape
        (default None).
    integer : bool, optional
        Whether to use an integer array. By default, float array is used
        (default False).

    Returns
    -------
    numpy.array
        The newly created shared numpy array.

    Raises
    ------
    ValueError
        If the input `data` (if provided) size is not equal to the created
        array.

    """
    size = np.prod(shape)

    if integer:
        array = Array(ctypes.c_int64, int(size))
        np_array = np.frombuffer(array.get_obj(), dtype="int64")
    else:
        array = Array(ctypes.c_double, int(size))
        np_array = np.frombuffer(array.get_obj())

    np_array = np_array.reshape(shape)

    if data is not None:
        if len(shape) != len(data.shape):
            raise ValueError("`data` must have the same shape as the created array.")

        same = all(x == y for x, y in zip(shape, data.shape))

        if not same:
            raise ValueError("`data` must have the same shape as the created array.")

        np_array[:] = data

    return np_array


class HRSampler(ABC):
    """
    The abstract base class for hit-and-run samplers.

    New samplers should derive from this class where possible to provide
    a uniform interface.

    Parameters
    ----------
    model : cobra.Model
        The cobra model from which to generate samples.
    thinning : int
        The thinning factor of the generated sampling chain. A thinning of
        10 means samples are returned every 10 steps.
    nproj : int > 0, optional
        How often to reproject the sampling point into the feasibility
        space. Avoids numerical issues at the cost of lower sampling. If
        you observe many equality constraint violations with
        `sampler.validate` you should lower this number (default None).
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp
        if None (default None).

    Attributes
    ----------
    feasibility_tol: float
        The tolerance used for checking equalities feasibility.
    bounds_tol: float
        The tolerance used for checking bounds feasibility.
    n_samples : int
        The total number of samples that have been generated by this
        sampler instance.
    retries : int
        The overall of sampling retries the sampler has observed. Larger
        values indicate numerical instabilities.
    problem : Problem
        A NamedTuple whose attributes define the entire sampling problem in
        matrix form.
    warmup : numpy.matrix
        A numpy matrix with as many columns as reactions in the model and
        more than 3 rows containing a warmup sample in each row. None if no
        warmup points have been generated yet.
    fwd_idx : numpy.array
        A numpy array having one entry for each reaction in the model,
        containing the index of the respective forward variable.
    rev_idx : numpy.array
        A numpy array having one entry for each reaction in the model,
        containing the index of the respective reverse variable.

    """

    def __init__(
        self,
        model: "Model",
        thinning: int,
        nproj: Optional[int] = None,
        seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Initialize a new sampler object.

        Other Parameters
        ----------------
        kwargs :
            Further keyword arguments are passed on to the parent class.

        Raises
        ------
        TypeError
            If integer problem is found.

        """
        # This currently has to be done to reset the solver basis which is
        # required to get deterministic warmup point generation
        # (in turn required for a working `seed`)
        if model.solver.is_integer:
            raise TypeError("Sampling does not work with integer problems.")

        self.model = model.copy()
        self.feasibility_tol = model.tolerance
        self.bounds_tol = model.tolerance
        self.thinning = thinning

        if nproj is None:
            self.nproj = int(min(len(self.model.variables) ** 3, 1e6))
        else:
            self.nproj = nproj

        self.n_samples = 0
        self.retries = 0
        self.problem = self.__build_problem()

        # Set up a map from reaction -> forward/reverse variable
        var_idx = {v: idx for idx, v in enumerate(self.model.variables)}

        self.fwd_idx = np.array(
            [var_idx[r.forward_variable] for r in self.model.reactions]
        )
        self.rev_idx = np.array(
            [var_idx[r.reverse_variable] for r in self.model.reactions]
        )
        self.warmup = None

        if seed is None:
            self._seed = int(time())
        else:
            self._seed = seed

        # Avoid overflow
        self._seed = self._seed % np.iinfo(np.int32).max

    def __build_problem(self) -> Problem:
        """Build the matrix representation of the sampling problem.

        Returns
        -------
        Problem
            The matrix representation in the form of a NamedTuple.

        """
        # Set up the mathematical problem
        prob = constraint_matrices(self.model, zero_tol=self.feasibility_tol)

        # check if there any non-zero equality constraints
        equalities = prob.equalities
        b = prob.b
        bounds = np.atleast_2d(prob.bounds).T
        var_bounds = np.atleast_2d(prob.variable_bounds).T
        homogeneous = all(np.abs(b) < self.feasibility_tol)
        fixed_non_zero = np.abs(prob.variable_bounds[:, 1]) > self.feasibility_tol
        fixed_non_zero &= prob.variable_fixed

        # check if there are any non-zero fixed variables, add them as
        # equalities to the stoichiometric matrix
        if any(fixed_non_zero):
            n_fixed = fixed_non_zero.sum()
            rows = np.zeros((n_fixed, prob.equalities.shape[1]))
            rows[range(n_fixed), np.where(fixed_non_zero)] = 1.0
            equalities = np.vstack([equalities, rows])
            var_b = prob.variable_bounds[:, 1]
            b = np.hstack([b, var_b[fixed_non_zero]])
            homogeneous = False

        # Set up a projection that can cast point into the nullspace
        nulls = nullspace(equalities)

        # convert bounds to a matrix and add variable bounds as well
        return Problem(
            equalities=shared_np_array(equalities.shape, equalities),
            b=shared_np_array(b.shape, b),
            inequalities=shared_np_array(prob.inequalities.shape, prob.inequalities),
            bounds=shared_np_array(bounds.shape, bounds),
            variable_fixed=shared_np_array(
                prob.variable_fixed.shape, prob.variable_fixed, integer=True
            ),
            variable_bounds=shared_np_array(var_bounds.shape, var_bounds),
            nullspace=shared_np_array(nulls.shape, nulls),
            homogeneous=homogeneous,
        )

    def generate_fva_warmup(self) -> None:
        """Generate the warmup points for the sampler.

        Generates warmup points by setting each flux as the sole objective
        and minimizing/maximizing it. Also caches the projection of the
        warmup points into the nullspace for non-homogeneous problems (only
        if necessary).

        Raises
        ------
        ValueError
            If flux cone contains a single point or the problem is
            inhomogeneous.

        """
        self.n_warmup = 0
        reactions = self.model.reactions
        self.warmup = np.zeros((2 * len(reactions), len(self.model.variables)))
        self.model.objective = Zero

        for sense in ("min", "max"):
            self.model.objective_direction = sense

            for i, r in enumerate(reactions):
                variables = (
                    self.model.variables[self.fwd_idx[i]],
                    self.model.variables[self.rev_idx[i]],
                )

                # Omit fixed reactions if they are non-homogeneous
                if r.upper_bound - r.lower_bound < self.bounds_tol:
                    logger.info(f"Skipping fixed reaction {r.id}")
                    continue

                self.model.objective.set_linear_coefficients(
                    {variables[0]: 1, variables[1]: -1}
                )

                self.model.slim_optimize()

                if not self.model.solver.status == OPTIMAL:
                    logger.info(f"Cannot maximize reaction {r.id}, skipping it.")
                    continue

                primals = self.model.solver.primal_values
                sol = [primals[v.name] for v in self.model.variables]
                self.warmup[
                    self.n_warmup,
                ] = sol
                self.n_warmup += 1

                # Reset objective
                self.model.objective.set_linear_coefficients(
                    {variables[0]: 0, variables[1]: 0}
                )

        # Shrink to measure
        self.warmup = self.warmup[0 : self.n_warmup, :]

        # Remove redundant search directions
        keep = np.logical_not(self._is_redundant(self.warmup))
        self.warmup = self.warmup[keep, :]
        self.n_warmup = self.warmup.shape[0]

        # Catch some special cases
        if len(self.warmup.shape) == 1 or self.warmup.shape[0] == 1:
            raise ValueError("Flux cone only consists a single point.")
        elif self.n_warmup == 2:
            if not self.problem.homogeneous:
                raise ValueError(
                    "Cannot sample from an inhomogenous problem "
                    "with only 2 search directions."
                )
            logger.info("All search directions on a line, adding another one.")
            newdir = self.warmup.T.dot([0.25, 0.25])
            self.warmup = np.vstack([self.warmup, newdir])
            self.n_warmup += 1

        # Shrink warmup points to measure
        self.warmup = shared_np_array(
            (self.n_warmup, len(self.model.variables)), self.warmup
        )

    def _reproject(self, p: np.ndarray) -> np.ndarray:
        """Reproject a point into the feasibility region.

        This function is guaranteed to return a new feasible point. However,
        no guarantee can be made in terms of proximity to the original
        point.

        Parameters
        ----------
        p : numpy.array
            The current sample point.

        Returns
        -------
        numpy.array
            A new feasible point. If `p` is feasible, it will return `p`.

        """
        nulls = self.problem.nullspace
        equalities = self.problem.equalities

        # don't reproject if point is feasible
        if np.allclose(
            equalities.dot(p), self.problem.b, rtol=0, atol=self.feasibility_tol
        ):
            new = p
        else:
            logger.info(
                f"Feasibility violated in sample {self.n_samples}, trying to reproject."
            )
            new = nulls.dot(nulls.T.dot(p))

        # Projections may violate bounds
        # set to random point in space in that case
        if any(new != p):
            logger.info(
                f"Re-projection failed in sample {self.n_samples}, "
                "using random point in space."
            )
            new = self._random_point()

        return new

    def _random_point(self) -> np.ndarray:
        """Find an approximately random point in the flux cone."""
        idx = np.random.randint(
            self.n_warmup, size=min(2, np.ceil(np.sqrt(self.n_warmup)))
        )
        return self.warmup[idx, :].mean(axis=0)

    def _is_redundant(self, matrix: np.matrix, cutoff: Optional[float] = None) -> bool:
        """Identify redundant rows in a matrix that can be removed."""
        cutoff = 1.0 - self.feasibility_tol
        # Avoid zero variances
        extra_col = matrix[:, 0] + 1
        # Avoid zero rows being correlated with constant rows
        extra_col[matrix.sum(axis=1) == 0] = 2
        corr = np.corrcoef(np.c_[matrix, extra_col])
        corr = np.tril(corr, -1)

        return (np.abs(corr) > cutoff).any(axis=1)

    def _bounds_dist(self, p: np.ndarray) -> np.ndarray:
        """Get the lower and upper bound distances. Negative is bad."""
        prob = self.problem
        lb_dist = (
            p
            - prob.variable_bounds[
                0,
            ]
        ).min()
        ub_dist = (
            prob.variable_bounds[
                1,
            ]
            - p
        ).min()

        if prob.bounds.shape[0] > 0:
            const = prob.inequalities.dot(p)
            const_lb_dist = (
                const
                - prob.bounds[
                    0,
                ]
            ).min()
            const_ub_dist = (
                prob.bounds[
                    1,
                ]
                - const
            ).min()
            lb_dist = min(lb_dist, const_lb_dist)
            ub_dist = min(ub_dist, const_ub_dist)

        return np.array([lb_dist, ub_dist])

    @abstractmethod
    def sample(self, n: int, fluxes: bool = True) -> pd.DataFrame:
        """Abstract sampling function.

        Should be overwritten by child classes.

        Parameters
        ----------
        n : int
            The number of samples that are generated at once.
        fluxes : bool, optional
            Whether to return fluxes or the internal solver variables. If
            set to False, will return a variable for each forward and
            backward flux as well as all additional variables you might
            have defined in the model (default True).

        Returns
        -------
        pandas.DataFrame
            Returns a pandas DataFrame with `n` rows, each containing a
            flux sample.

        """
        raise NotImplementedError(
            "This method needs to be implemented by the subclass."
        )

    def batch(
        self, batch_size: int, batch_num: int, fluxes: bool = True
    ) -> pd.DataFrame:
        """Create a batch generator.

        This is useful to generate `batch_num` batches of `batch_size`
        samples each.

        Parameters
        ----------
        batch_size : int
            The number of samples contained in each batch.
        batch_num : int
            The number of batches in the generator.
        fluxes : bool, optional
            Whether to return fluxes or the internal solver variables. If
            set to False, will return a variable for each forward and
            backward flux as well as all additional variables you might
            have defined in the model (default True).

        Yields
        ------
        pandas.DataFrame
            A DataFrame with dimensions (batch_size x n_r) containing
            a valid flux sample for a total of n_r reactions (or variables
            if fluxes=False) in each row.

        """
        for _ in range(batch_num):
            yield self.sample(batch_size, fluxes=fluxes)

    def validate(self, samples: np.matrix) -> np.ndarray:
        """Validate a set of samples for equality and inequality feasibility.

        Can be used to check whether the generated samples and warmup points
        are feasible.

        Parameters
        ----------
        samples : numpy.matrix
            Must be of dimension (samples x n_reactions). Contains the
            samples to be validated. Samples must be from fluxes.

        Returns
        -------
        numpy.array
            A one-dimensional numpy array containing
            a code of 1 to 3 letters denoting the validation result:
            - 'v' means feasible in bounds and equality constraints
            - 'l' means a lower bound violation
            - 'u' means a lower bound validation
            - 'e' means and equality constraint violation

        Raises
        ------
        ValueError
            If wrong number of columns.

        """
        samples = np.atleast_2d(samples)
        prob = self.problem

        if samples.shape[1] == len(self.model.reactions):
            S = create_stoichiometric_matrix(self.model)
            b = np.array(
                [self.model.constraints[m.id].lb for m in self.model.metabolites]
            )
            bounds = np.array([r.bounds for r in self.model.reactions]).T
        elif samples.shape[1] == len(self.model.variables):
            S = prob.equalities
            b = prob.b
            bounds = prob.variable_bounds
        else:
            raise ValueError(
                "Wrong number of columns. Samples must have a "
                "column for each flux or variable defined in the "
                "model."
            )

        feasibility = np.abs(S.dot(samples.T).T - b).max(axis=1)
        lb_error = (
            samples
            - bounds[
                0,
            ]
        ).min(axis=1)
        ub_error = (
            bounds[
                1,
            ]
            - samples
        ).min(axis=1)

        if samples.shape[1] == len(self.model.variables) and prob.inequalities.shape[0]:
            consts = prob.inequalities.dot(samples.T)
            lb_error = np.minimum(
                lb_error,
                (
                    consts
                    - prob.bounds[
                        0,
                    ]
                ).min(axis=1),
            )
            ub_error = np.minimum(
                ub_error,
                (
                    prob.bounds[
                        1,
                    ]
                    - consts
                ).min(axis=1),
            )

        valid = (
            (feasibility < self.feasibility_tol)
            & (lb_error > -self.bounds_tol)
            & (ub_error > -self.bounds_tol)
        )
        codes = np.repeat("", valid.shape[0]).astype(np.dtype((str, 3)))
        codes[valid] = "v"
        codes[lb_error <= -self.bounds_tol] = np.char.add(
            codes[lb_error <= -self.bounds_tol], "l"
        )
        codes[ub_error <= -self.bounds_tol] = np.char.add(
            codes[ub_error <= -self.bounds_tol], "u"
        )
        codes[feasibility > self.feasibility_tol] = np.char.add(
            codes[feasibility > self.feasibility_tol], "e"
        )

        return codes



"""General utilities used across the package."""

import textwrap
from typing import Any

from depinfo import print_dependencies


def format_long_string(string: str, max_length: int = 50) -> str:
    """Shorten long string into a small string with ellipsis.

    Parameters
    ----------
    string: str
        The long string to shorten.
    max_length: int, optional
        The maximum length after which to append ellipsis (default 50).

    Returns
    -------
    str
        The shortened string.

    """
    return textwrap.shorten(string, width=max_length, placeholder="...")


class AutoVivification(dict):
    """
    Implementation of Perl's autovivification feature.

    Notes
    -----
    For more information, check https://stackoverflow.com/a/652284/280182 .

    """

    def __getitem__(self, item: Any) -> Any:
        """Retrieve if item is found, else add it.

        Parameters
        ----------
        item: Any
            The object to look for.

        Returns
        -------
        Any
            The retrieved object.

        """
        try:
            value = super().__getitem__(item)
        except KeyError:
            value = self[item] = type(self)()

        return value


def show_versions() -> None:
    """Print dependency information."""
    print_dependencies("cobra")



"""Context manager for the package."""

from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Optional


if TYPE_CHECKING:
    from cobra import Object


class HistoryManager:
    """
    Define a base context manager.

    It records a list of actions to be taken at a later time.
    This is used to implement context managers that allow temporary
    changes to a `cobra.core.Model`.

    """

    def __init__(self, **kwargs) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)
        # this acts like a stack
        self._history = []

    def __call__(self, operation: Callable[[Any], Any]) -> None:
        """Add the corresponding operation to the history stack.

        Parameters
        ----------
        operation : callable
            A function to be called at a later time.

        """
        self._history.append(operation)

    def reset(self) -> None:
        """Trigger executions for all items in the stack in reverse order."""
        while self._history:
            entry = self._history.pop()
            entry()

    def size(self) -> int:
        """Calculate number of operations on the stack."""
        return len(self._history)


def get_context(obj: "Object") -> Optional[HistoryManager]:
    """Search for a context manager.

    Parameters
    ----------
    obj: cobra.Object
        The cobra.Object for which to search context manager.

    Returns
    -------
    HistoryManager or None
        HistoryManager instance, or None if no context manager is found.

    Raises
    ------
    AttributeError
        If no context manager is found.
    IndexError
        If no context manager is found.

    """
    # works for cobra.core.Model objects
    try:
        return obj._contexts[-1]
    except (AttributeError, IndexError):
        pass
    # works for objects other than cobra.core.Model
    try:
        return obj._model._contexts[-1]
    except (AttributeError, IndexError):
        pass


def resettable(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """
    Simplify the context management of simple object attributes.

    It gets the value of the attribute prior to setting it, and stores
    a function to set the value to the old value in the
    `cobra.util.HistoryManager`.

    Parameters
    ----------
    func: callable
        The function to decorate.

    Returns
    -------
    callable
        The decorated function.

    """

    def wrapper(self, new_value):
        context = get_context(self)
        if context:
            old_value = getattr(self, func.__name__)
            # Don't clutter the context with unchanged variables
            if old_value == new_value:
                return
            context(partial(func, self, old_value))

        func(self, new_value)

    return wrapper



"""Provide functions for pruning reactions, metabolites and genes."""

from ast import And, Module, NodeTransformer
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple
from warnings import warn

from ..core.gene import ast2str, eval_gpr, parse_gpr


if TYPE_CHECKING:
    from cobra import Gene, Metabolite, Model, Reaction


def prune_unused_metabolites(model: "Model") -> Tuple["Model", List["Metabolite"]]:
    """Remove metabolites not involved in any reactions.

    Parameters
    ----------
    model: cobra.Model
        The model to remove unused metabolites from.

    Returns
    -------
    cobra.Model
        Input model with unused metabolites removed.
    list of cobra.Metabolite
        List of metabolites that were removed.

    """
    output_model = model.copy()
    inactive_metabolites = [
        m for m in output_model.metabolites if len(m.reactions) == 0
    ]
    output_model.remove_metabolites(inactive_metabolites)
    return output_model, inactive_metabolites


def prune_unused_reactions(model: "Model") -> Tuple["Model", List["Reaction"]]:
    """Remove reactions with no assigned metabolites, returns pruned model.

    Parameters
    ----------
    model: cobra.Model
        The model to remove unused reactions from.

    Returns
    -------
    cobra.Model
        Input model with unused reactions removed.
    list of cobra.Reaction
        List of reactions that were removed.

    """
    output_model = model.copy()
    reactions_to_prune = [r for r in output_model.reactions if len(r.metabolites) == 0]
    output_model.remove_reactions(reactions_to_prune)
    return output_model, reactions_to_prune


def undelete_model_genes(model: "Model") -> None:
    """Undo the effects of a call to `delete_model_genes` in place.

    Parameters
    ----------
    model: cobra.Model
        The model which will be modified in place.

    """
    if model._trimmed_genes is not None:
        for x in model._trimmed_genes:
            x.functional = True

    if model._trimmed_reactions is not None:
        for (
            the_reaction,
            (lower_bound, upper_bound),
        ) in model._trimmed_reactions.items():
            the_reaction.lower_bound = lower_bound
            the_reaction.upper_bound = upper_bound

    model._trimmed_genes = []
    model._trimmed_reactions = {}
    model._trimmed = False


def get_compiled_gene_reaction_rules(model: "Model") -> Dict["Reaction", Module]:
    """Generate a dictionary of compiled gene-reaction rules.

    Any gene-reaction rule expression which cannot be compiled or do not
    evaluate after compiling will be excluded. The result can be used in the
    `find_gene_knockout_reactions` function to speed up evaluation of these
    rules.

    Parameters
    ----------
    model: cobra.Model
        The model to get gene-reaction rules for.

    Returns
    -------
    dict of cobra.Reaction, ast.Module
        The dictionary of cobra.Reaction objects as keys and ast.Module
        objects as keys.

    """
    return {r: parse_gpr(r.gene_reaction_rule)[0] for r in model.reactions}


def find_gene_knockout_reactions(
    model: "Model",
    gene_list: List["Gene"],
    compiled_gene_reaction_rules: Optional[Dict["Reaction", Module]] = None,
) -> List["Reaction"]:
    """Identify reactions which will be disabled when genes are knocked out.

    Parameters
    ----------
    model: cobra.Model
        The model for which to find gene knock-out reactions.
    gene_list: list of cobra.Gene
        The list of genes to knock-out.
    compiled_gene_reaction_rules: dict of {reaction: compiled_string},
                                  optional
        If provided, this gives pre-compiled gene-reaction rule strings.
        The compiled rule strings can be evaluated much faster. If a rule
        is not provided, the regular expression evaluation will be used.
        Because not all gene-reaction rule strings can be evaluated, this
        dict must exclude any rules which can not be used with eval
        (default None).

    Returns
    -------
    list of cobra.Reaction
       The list of cobra.Reaction objects which will be disabled.

    .. deprecated:: 0.22.1
        Internal function that has outlived its purpose.

    """
    warn(
        "The function `find_gene_knockout_reactions` has outlived its purpose. "
        "It will be removed in the next minor version (0.23.0).",
        DeprecationWarning,
    )
    potential_reactions = set()
    for gene in gene_list:
        if isinstance(gene, str):
            gene = model.genes.get_by_id(gene)
        potential_reactions.update(gene._reaction)
    gene_set = {str(i) for i in gene_list}
    if compiled_gene_reaction_rules is None:
        compiled_gene_reaction_rules = {
            r: parse_gpr(r.gene_reaction_rule)[0] for r in potential_reactions
        }

    return [
        r
        for r in potential_reactions
        if not eval_gpr(compiled_gene_reaction_rules[r], gene_set)
    ]


def delete_model_genes(
    model: "Model",
    gene_list: List["Gene"],
    cumulative_deletions: bool = True,
    disable_orphans: bool = False,
) -> None:
    """Temporarily remove the effect of genes in `gene_list`.

    It sets the bounds to "zero" for reactions catalysed by the genes in
    `gene_list` if deleting the genes stops the reactions from proceeding.

    Parameters
    ----------
    model: cobra.Model
        The model whose reaction bounds are to be set.
    gene_list: list of cobra.Gene
        The list of genes to knock-out.
    cumulative_deletions: bool, optional
        If True, then any previous deletions will be maintained in the
        model (default True).
    disable_orphans: bool, optional
        If True, then orphan reactions will be disabled. Currently, this
        is not implemented (default False).

    """
    if disable_orphans:
        raise NotImplementedError("disable_orphans not implemented")
    if not hasattr(model, "_trimmed"):
        model._trimmed = False
        model._trimmed_genes = []
        model._trimmed_reactions = {}  # Store the old bounds in here.
    # older models have this
    if model._trimmed_genes is None:
        model._trimmed_genes = []
    if model._trimmed_reactions is None:
        model._trimmed_reactions = {}
    # Allow a single gene to be fed in as a string instead of a list.
    if not hasattr(gene_list, "__iter__") or hasattr(
        gene_list, "id"
    ):  # cobra.Gene has __iter__
        gene_list = [gene_list]

    if not hasattr(gene_list[0], "id"):
        if gene_list[0] in model.genes:
            tmp_gene_dict = dict([(x.id, x) for x in model.genes])
        else:
            # assume we're dealing with names if no match to an id
            tmp_gene_dict = dict([(x.name, x) for x in model.genes])
        gene_list = [tmp_gene_dict[x] for x in gene_list]

    # Make the genes non-functional
    for x in gene_list:
        x.functional = False

    if cumulative_deletions:
        gene_list.extend(model._trimmed_genes)
    else:
        undelete_model_genes(model)

    for the_reaction in find_gene_knockout_reactions(model, gene_list):
        # Running this on an already deleted reaction will overwrite the
        # stored reaction bounds.
        if the_reaction in model._trimmed_reactions:
            continue
        old_lower_bound = the_reaction.lower_bound
        old_upper_bound = the_reaction.upper_bound
        model._trimmed_reactions[the_reaction] = (
            old_lower_bound,
            old_upper_bound,
        )
        the_reaction.lower_bound = 0.0
        the_reaction.upper_bound = 0.0
        model._trimmed = True

    model._trimmed_genes = list(set(model._trimmed_genes + gene_list))


class _GeneRemover(NodeTransformer):
    """
    Class to represent a gene set remover.

    Parameters
    ----------
    target_genes: list of cobra.Gene
        A list of genes to be removed.

    """

    def __init__(self, target_genes: List["Gene"], **kwargs) -> None:
        """Initialize a new object.

        Other Parameters
        ----------------
        kwargs:
            Further keyword arguments are passed on to the parent class.

        """
        super().__init__(**kwargs)
        self.target_genes = {str(i) for i in target_genes}

    def visit_Name(self, node: "Gene") -> Optional["Gene"]:
        """Remove a gene.

        Parameters
        ----------
        node: cobra.Gene
            The gene to remove.

        Returns
        -------
        cobra.Gene or None
            None if gene object is in `target_genes`.

        """
        return None if node.id in self.target_genes else node

    def visit_BoolOp(self, node: "Gene") -> Optional["Gene"]:
        """Rules for boolean operations.

        Parameters
        ----------
        node: cobra.Gene
            The gene to apply rules to.

        Returns
        -------
        cobra.Gene or None
            None if size of node values is less after applying rule.

        """
        original_n = len(node.values)
        self.generic_visit(node)
        if len(node.values) == 0:
            return None
        # AND with any entities removed
        if len(node.values) < original_n and isinstance(node.op, And):
            return None
        # if one entity in an OR was removed, just that entity passed up
        if len(node.values) == 1:
            return node.values[0]
        return node


def remove_genes(
    model: "Model", gene_list: List["Gene"], remove_reactions: bool = True
) -> None:
    """Remove genes entirely from the model.

    This will also simplify all gene-reaction rules with the genes
    inactivated.

    Parameters
    ----------
    model: cobra.Model
        The model to remove genes from.
    gene_list: list of cobra.Gene
        The list of gene objects to remove.
    remove_reactions: bool, optional
        Whether to remove reactions associated with genes in `gene_list`
        (default True).

    """
    gene_set = {model.genes.get_by_id(str(i)) for i in gene_list}
    gene_id_set = {i.id for i in gene_set}
    remover = _GeneRemover(gene_id_set)
    ast_rules = get_compiled_gene_reaction_rules(model)
    target_reactions = []
    for reaction, rule in ast_rules.items():
        if reaction.gene_reaction_rule is None or len(reaction.gene_reaction_rule) == 0:
            continue
        # reactions to remove
        if remove_reactions and not eval_gpr(rule, gene_id_set):
            target_reactions.append(reaction)
        else:
            # if the reaction is not removed, remove the gene
            # from its gpr
            remover.visit(rule)
            new_rule = ast2str(rule)
            if new_rule != reaction.gene_reaction_rule:
                reaction.gene_reaction_rule = new_rule
    for gene in gene_set:
        model.genes.remove(gene)
        # remove reference to the gene in all groups
        associated_groups = model.get_associated_groups(gene)
        for group in associated_groups:
            group.remove_members(gene)
    model.remove_reactions(target_reactions)
